"""
Contains Validator classes which each have their own error to handle and mark.
"""
import logging
from typing import Any, List, Optional, Set, Tuple, Type, Union

import geopandas as gpd
import numpy as np
from shapely.affinity import scale
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from shapely.geometry.base import BaseGeometry
from shapely.ops import linemerge, split

from fractopo.general import (
    compare_unit_vector_orientation,
    create_unit_vector,
    determine_node_junctions,
    get_trace_coord_points,
    get_trace_endpoints,
)
from fractopo.tval.trace_validation_utils import (
    is_underlapping,
    segment_within_buffer,
    split_to_determine_triangle_errors,
)

log = logging.getLogger(__name__)


class BaseValidator:

    """
    Base validator that all classes inherit.

    TODO: Validation requires overhaul at some point.
    """

    ERROR = "BASE ERROR"
    INTERACTION_NODES_COLUMN = "IP"
    LINESTRING_ONLY = True

    @staticmethod
    def fix_method(**_) -> Optional[LineString]:
        """
        Abstract fix method.
        """
        raise NotImplementedError("fix_method not implemented.")

    @staticmethod
    def validation_method(**_) -> bool:
        """
        Abstract validation method.
        """
        raise NotImplementedError("validation_method not implemented.")


class GeomTypeValidator(BaseValidator):

    """
    Validates the geometry type.

    Validates that all traces are LineStrings. Tries to use shapely.ops.linemerge
    to merge MultiLineStrings into LineStrings.
    """

    ERROR = "GEOM TYPE MULTILINESTRING"
    LINESTRING_ONLY = False

    @staticmethod
    def fix_method(geom: Any, **_) -> Optional[LineString]:
        """
        Fix mergeable MultiLineStrings to LineStrings.

        E.g. mergeable MultiLineString

        >>> mls = MultiLineString([((0, 0), (1, 1)), ((1, 1), (2, 2))])
        >>> GeomTypeValidator.fix_method(geom=mls).wkt
        'LINESTRING (0 0, 1 1, 2 2)'

        Unhandled types will just return as None.

        >>> GeomTypeValidator.fix_method(geom=Point(1, 1)) is None
        True

        """
        if isinstance(geom, LineString):
            return geom
        if not isinstance(geom, MultiLineString):
            return None
        fixed_geom = linemerge(geom)
        return fixed_geom if isinstance(fixed_geom, LineString) else None

    @staticmethod
    def validation_method(geom: Any, **_) -> bool:
        """
        Validate geometries.

        E.g. Anything but LineString

        >>> GeomTypeValidator.validation_method(Point(1, 1))
        False

        With LineString:

        >>> GeomTypeValidator.validation_method(LineString([(0, 0), (1, 1)]))
        True

        """
        return isinstance(geom, LineString)


class MultiJunctionValidator(BaseValidator):

    """
    Validates that junctions consists of a maximum of two lines crossing.
    """

    ERROR = "MULTI JUNCTION"

    @staticmethod
    def validation_method(idx: int, faulty_junctions: Set[int], **_) -> bool:
        """
        Validate for multi junctions between traces.

        >>> MultiJunctionValidator.validation_method(
        ...     idx=1, faulty_junctions=set([1, 2])
        ... )
        False

        """
        return idx not in faulty_junctions

    @staticmethod
    def determine_faulty_junctions(
        all_nodes: List[Tuple[Point, ...]],
        snap_threshold: float,
        snap_threshold_error_multiplier: float,
    ) -> Set[int]:
        """
        Determine when a point of interest represents a multi junction.

        I.e. when there are more than 2 points within the buffer distance of
        each other.

        Two points is the limit because an intersection point between two traces
        is added to the list of interest points twice, once for each trace.

        Faulty junction can also represent a slightly overlapping trace i.e.
        a snapping error.

        >>> all_nodes = [
        ...     (Point(0, 0), Point(1, 1)),
        ...     (Point(1, 1),),
        ...     (Point(5, 5),),
        ...     (Point(0, 0), Point(1, 1)),
        ... ]
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> MultiJunctionValidator.determine_faulty_junctions(
        ...     all_nodes, snap_threshold, snap_threshold_error_multiplier
        ... )
        {0, 1, 3}

        """
        return determine_node_junctions(
            nodes=all_nodes,
            snap_threshold=snap_threshold,
            snap_threshold_error_multiplier=snap_threshold_error_multiplier,
            error_threshold=2,
        )


class VNodeValidator(BaseValidator):

    """
    Finds V-nodes within trace data.
    """

    ERROR = "V NODE"

    @staticmethod
    def validation_method(idx: int, vnodes: Set[int], **_) -> bool:
        """
        Validate for V-nodes.

        >>> VNodeValidator.validation_method(idx=1, vnodes=set([1, 2]))
        False

        >>> VNodeValidator.validation_method(idx=5, vnodes=set([1, 2]))
        True

        """
        return idx not in vnodes

    @staticmethod
    def determine_v_nodes(
        endpoint_nodes: List[Tuple[Point, ...]],
        snap_threshold: float,
        snap_threshold_error_multiplier: float,
    ) -> Set[int]:
        """
        Determine V-node errors.

        >>> endpoint_nodes = [
        ...     (Point(0, 0), Point(1, 1)),
        ...     (Point(1, 1),),
        ... ]
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> VNodeValidator.determine_v_nodes(
        ...     endpoint_nodes, snap_threshold, snap_threshold_error_multiplier
        ... )
        {0, 1}

        """
        return determine_node_junctions(
            nodes=endpoint_nodes,
            snap_threshold=snap_threshold,
            snap_threshold_error_multiplier=snap_threshold_error_multiplier,
            error_threshold=1,
        )


class MultipleCrosscutValidator(BaseValidator):

    """
    Find traces that cross-cut each other multiple times.

    This also indicates the possibility of duplicate traces.
    """

    ERROR = "MULTIPLE CROSSCUTS"

    @staticmethod
    def validation_method(
        geom: LineString,
        trace_candidates: gpd.GeoSeries,
        **_,
    ) -> bool:
        """
        Validate for multiple crosscuts.

        >>> geom = LineString([Point(-3, -4), Point(-3, -1)])
        >>> trace_candidates = gpd.GeoSeries(
        ...     [
        ...         LineString(
        ...             [Point(-4, -3), Point(-2, -3), Point(-4, -2), Point(-2, -1)]
        ...         )
        ...     ]
        ... )
        >>> MultipleCrosscutValidator.validation_method(geom, trace_candidates)
        False

        >>> geom = LineString([Point(-3, -4), Point(-3, -4.5)])
        >>> trace_candidates = gpd.GeoSeries(
        ...     [
        ...         LineString(
        ...             [Point(-4, -3), Point(-2, -3), Point(-4, -2), Point(-2, -1)]
        ...         )
        ...     ]
        ... )
        >>> MultipleCrosscutValidator.validation_method(geom, trace_candidates)
        True

        """
        if not any(trace_candidates.intersects(geom)):
            return True
        intersection_geoms = trace_candidates.intersection(geom)
        return not any(
            len(list(geom.geoms)) > 2
            for geom in intersection_geoms
            if isinstance(geom, MultiPoint)
        )


class UnderlappingSnapValidator(BaseValidator):

    """
    Find snapping errors of underlapping traces.

    Uses a multiple of the given snap_threshold.
    """

    ERROR = "UNDERLAPPING SNAP"
    _OVERLAPPING = "OVERLAPPING SNAP"
    _UNDERLAPPING = "UNDERLAPPING SNAP"

    @classmethod
    def validation_method(
        cls,
        geom: LineString,
        trace_candidates: gpd.GeoSeries,
        snap_threshold: float,
        snap_threshold_error_multiplier: float,
        **_,
    ) -> bool:
        """
        Validate for UnderlappingSnapValidator errors.

        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> geom = LineString(
        ...     [
        ...         (0, 0),
        ...         (0, 1 + snap_threshold * snap_threshold_error_multiplier * 0.99),
        ...     ]
        ... )
        >>> trace_candidates = gpd.GeoSeries([LineString([(-1, 1), (1, 1)])])
        >>> UnderlappingSnapValidator.validation_method(
        ...     geom, trace_candidates, snap_threshold, snap_threshold_error_multiplier
        ... )
        False

        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> geom = LineString([(0, 0), (0, 1)])
        >>> trace_candidates = gpd.GeoSeries([LineString([(-1, 1), (1, 1)])])
        >>> UnderlappingSnapValidator.validation_method(
        ...     geom, trace_candidates, snap_threshold, snap_threshold_error_multiplier
        ... )
        True

        """
        if len(trace_candidates) == 0:
            return True

        endpoints = get_trace_endpoints(geom)
        for endpoint in endpoints:
            # Check that if endpoint is well snapped to one trace another trace
            # won't have chance to cause a validation error (false positive).
            if any(trace_candidates.distance(endpoint) < snap_threshold):
                continue
            trace: LineString
            for trace in trace_candidates.geometry.values:
                if (
                    snap_threshold
                    < trace.distance(endpoint)
                    < snap_threshold * snap_threshold_error_multiplier
                ):
                    # Determine if its underlappings or overlapping
                    is_ul_result = is_underlapping(
                        geom,
                        trace,
                        endpoint,
                        snap_threshold,
                        snap_threshold_error_multiplier,
                    )
                    if is_ul_result is None:
                        raise ValueError("Expected is_ul_result to not be None.")
                    if is_ul_result:
                        # Underlapping
                        cls.ERROR = cls._UNDERLAPPING
                        return False
                    cls.ERROR = cls._OVERLAPPING
                    return False

        return True


class TargetAreaSnapValidator(BaseValidator):

    """
    Validator for traces that underlap the target area.
    """

    ERROR = "TRACE UNDERLAPS TARGET AREA"

    @staticmethod
    def validation_method(
        geom: LineString,
        area: gpd.GeoDataFrame,
        snap_threshold: float,
        snap_threshold_error_multiplier: float,
        area_edge_snap_multiplier: float,
        **_,
    ) -> bool:
        """
        Validate for trace underlaps.

        >>> geom = LineString([(0, 0), (0, 1)])
        >>> area = gpd.GeoDataFrame(
        ...     geometry=[Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])]
        ... )
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> area_edge_snap_multiplier = 1.0
        >>> TargetAreaSnapValidator.validation_method(
        ...     geom,
        ...     area,
        ...     snap_threshold,
        ...     snap_threshold_error_multiplier,
        ...     area_edge_snap_multiplier,
        ... )
        True

        >>> geom = LineString([(0.5, 0.5), (0.5, 0.98)])
        >>> area = gpd.GeoDataFrame(
        ...     geometry=[Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])]
        ... )
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> area_edge_snap_multiplier = 10
        >>> TargetAreaSnapValidator.validation_method(
        ...     geom,
        ...     area,
        ...     snap_threshold,
        ...     snap_threshold_error_multiplier,
        ...     area_edge_snap_multiplier,
        ... )
        False

        """
        endpoints = get_trace_endpoints(geom)
        for endpoint in endpoints:
            for area_polygon in area.geometry.values:
                if TargetAreaSnapValidator.is_candidate_underlapping(
                    endpoint,
                    geom,
                    area_polygon,
                    snap_threshold=snap_threshold,
                ):
                    # if endpoint.within(area_polygon) and geom.within(area_polygon):
                    # Point and trace completely within the area, does not
                    # intersect its edge.
                    if (
                        snap_threshold
                        <= endpoint.distance(area_polygon.boundary)
                        < snap_threshold
                        * snap_threshold_error_multiplier
                        * area_edge_snap_multiplier
                    ):
                        return False
        return True

    @staticmethod
    def simple_underlapping_checks(
        endpoint: Point,
        geom: LineString,
        area_polygon: Union[Polygon, MultiPolygon],
        snap_threshold: float,
    ) -> Optional[bool]:
        """
        Perform simple underlapping checks.
        """
        endpoint_within = endpoint.within(area_polygon)
        # If not even inside target area, not a candidate
        if not endpoint_within:
            return False

        # Easy case: endpoint within target area and trace completely inside
        # target area
        if endpoint_within and geom.within(area_polygon):
            return True

        # Does an affine scale to catch traces that intersect the target area
        # edge at their other endpoint. This check overlaps with previous but
        # shapely.affinity.scale is not completely 'stable'
        if endpoint_within and geom.within(
            scale(area_polygon, xfact=1 + snap_threshold, yfact=1 + snap_threshold)
        ):
            return True
        return None

    @staticmethod
    def is_candidate_underlapping(
        endpoint: Point,
        geom: LineString,
        area_polygon: Union[Polygon, MultiPolygon],
        snap_threshold: float,
    ) -> bool:
        """
        Determine if endpoint is candidate for Underlapping error.
        """
        is_simple_underlapping = TargetAreaSnapValidator.simple_underlapping_checks(
            endpoint=endpoint,
            geom=geom,
            area_polygon=area_polygon,
            snap_threshold=snap_threshold,
        )

        if is_simple_underlapping is not None:
            return is_simple_underlapping

        # Split trace with area polygon
        geom_split = list(split(geom, area_polygon).geoms)

        # If not splittable, not a candidate
        if len(geom_split) == 1:
            return False

        # Filter segments that do not intersect endpoint
        endpoint_segments = [
            segment
            for segment in geom_split
            if segment.distance(endpoint) < snap_threshold
        ]

        # More than 1 shouldn't intersect one endpoint
        if len(endpoint_segments) != 1:
            log.warning("Expected only one segment to be close to endpoint.")
            return False
        endpoint_segment = endpoint_segments[0]

        # Make sure result of split is LineString
        if not isinstance(endpoint_segment, LineString):
            log.warning("Expected endpoint_segment to be of type LineString.")
            return False

        # If the LineString intersects polygon boundary -> not a candidate
        if endpoint_segment.distance(area_polygon.boundary) < snap_threshold:
            return False
        return True


class GeomNullValidator(BaseValidator):

    """
    Validate the geometry for NULL GEOMETRY errors.
    """

    ERROR = "NULL GEOMETRY"
    LINESTRING_ONLY = False

    @staticmethod
    def validation_method(geom: Any, **_) -> bool:
        """
        Validate for empty and null geometries.

        E.g. some validations are handled by GeomTypeValidator

        >>> GeomNullValidator.validation_method(Point(1, 1))
        True

        Empty geometries are not valid.

        >>> GeomNullValidator.validation_method(LineString())
        False

        >>> GeomNullValidator.validation_method(LineString([(-1, 1), (1, 1)]))
        True

        """
        if not isinstance(geom, BaseGeometry):
            return False
        if geom is None:
            return False
        if geom.is_empty:
            return False
        return True


class StackedTracesValidator(BaseValidator):

    """
    Find stacked traces and small triangle intersections.
    """

    ERROR = "STACKED TRACES"

    @staticmethod
    def validation_method(
        geom: LineString,
        trace_candidates: gpd.GeoSeries,
        snap_threshold: float,
        snap_threshold_error_multiplier: float,
        overlap_detection_multiplier: float,
        triangle_error_snap_multiplier: float,
        **_,
    ) -> bool:
        """
        Validate for stacked traces and small triangle intersections.

        >>> geom = LineString([(0, 0), (0, 1)])
        >>> trace_candidates = gpd.GeoSeries([LineString([(0, -1), (0, 2)])])
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> overlap_detection_multiplier = 50
        >>> triangle_error_snap_multiplier = 10
        >>> StackedTracesValidator.validation_method(
        ...     geom,
        ...     trace_candidates,
        ...     snap_threshold,
        ...     snap_threshold_error_multiplier,
        ...     overlap_detection_multiplier,
        ...     triangle_error_snap_multiplier,
        ... )
        False

        >>> geom = LineString([(10, 0), (10, 1)])
        >>> trace_candidates = gpd.GeoSeries([LineString([(0, -1), (0, 2)])])
        >>> snap_threshold = 0.01
        >>> snap_threshold_error_multiplier = 1.1
        >>> overlap_detection_multiplier = 50
        >>> triangle_error_snap_multiplier = 10
        >>> StackedTracesValidator.validation_method(
        ...     geom,
        ...     trace_candidates,
        ...     snap_threshold,
        ...     snap_threshold_error_multiplier,
        ...     overlap_detection_multiplier,
        ...     triangle_error_snap_multiplier,
        ... )
        True

        """
        # Cannot overlap if nothing to overlap
        if len(trace_candidates) == 0:
            return True

        trace_candidates_multils = MultiLineString(
            [
                tc
                for tc in trace_candidates.geometry.values
                if isinstance(tc, LineString)
                and tc.buffer(
                    snap_threshold
                    * overlap_detection_multiplier
                    * snap_threshold_error_multiplier
                ).intersects(geom)
            ]
        )
        # Test for overlapping traces.
        if segment_within_buffer(
            geom,
            trace_candidates_multils,
            snap_threshold=snap_threshold,
            snap_threshold_error_multiplier=snap_threshold_error_multiplier,
            overlap_detection_multiplier=overlap_detection_multiplier,
        ):
            return False

        # Test for small triangles created by the intersection of two traces.
        splitter_trace: LineString
        for splitter_trace in trace_candidates.geometry.values:
            if split_to_determine_triangle_errors(
                geom,
                splitter_trace,
                snap_threshold=snap_threshold,
                triangle_error_snap_multiplier=triangle_error_snap_multiplier,
            ):
                return False

        return True


class SimpleGeometryValidator(BaseValidator):

    """
    Use shapely is_simple and is_ring attributes to validate LineString.

    Checks that trace does not cut itself.
    """

    ERROR = "CUTS ITSELF"

    @staticmethod
    def validation_method(geom: LineString, **_) -> bool:
        """
        Validate for self-intersections.
        """
        return geom.is_simple and not geom.is_ring


class SharpCornerValidator(BaseValidator):

    """
    Find sharp cornered traces.
    """

    ERROR = "SHARP TURNS"

    @staticmethod
    def validation_method(
        geom: LineString,
        sharp_avg_threshold: float,
        sharp_prev_seg_threshold: float,
        **_,
    ) -> bool:
        """
        Validate for sharp cornered traces.
        """
        geom_coords = get_trace_coord_points(geom)
        if len(geom_coords) == 2:
            # If LineString consists of two Points -> No sharp corners.
            return True
        trace_unit_vector = create_unit_vector(geom_coords[0], geom_coords[-1])
        if any(np.isnan(trace_unit_vector)):
            return False
        for idx, segment_start in enumerate(geom_coords):
            if idx == len(geom_coords) - 1:
                break
            segment_end: Point = geom_coords[idx + 1]
            segment_unit_vector = create_unit_vector(segment_start, segment_end)
            # Compare the two unit vectors
            if any(
                np.isnan(segment_unit_vector)
            ) or not compare_unit_vector_orientation(
                trace_unit_vector, segment_unit_vector, sharp_avg_threshold
            ):
                return False
            # Check how much of a change compared to previous segment.
            if idx != 0:
                # Cannot get previous if node is first
                prev_segment_unit_vector = create_unit_vector(
                    geom_coords[idx - 1], geom_coords[idx]
                )
                if any(
                    np.isnan(prev_segment_unit_vector)
                ) or not compare_unit_vector_orientation(
                    segment_unit_vector,
                    prev_segment_unit_vector,
                    sharp_prev_seg_threshold,
                ):
                    return False

        return True


class EmptyTargetAreaValidator(BaseValidator):

    """
    Stub validator for empty target area.

    Validation for this error occurs at the start of run_validation.
    """

    ERROR = "EMPTY TARGET AREA"


MINOR_VALIDATORS = (
    SimpleGeometryValidator,
    MultiJunctionValidator,
    VNodeValidator,
    MultipleCrosscutValidator,
    UnderlappingSnapValidator,
    TargetAreaSnapValidator,
    StackedTracesValidator,
    SharpCornerValidator,
)

MAJOR_VALIDATORS = (GeomNullValidator, GeomTypeValidator)

ALL_VALIDATORS = MAJOR_VALIDATORS + MINOR_VALIDATORS

MAJOR_ERRORS = (GeomTypeValidator.ERROR, GeomNullValidator.ERROR)
ValidatorClass = Type[BaseValidator]

VALIDATION_REQUIRES_NODES = (MultiJunctionValidator, VNodeValidator)



"""
Functions for extracting branches and nodes from trace maps.

branches_and_nodes is the main entrypoint.
"""
import logging
import math
from itertools import chain, compress
from typing import Dict, List, Optional, Tuple, Union

import geopandas as gpd
import numpy as np
from geopandas.sindex import PyGEOSSTRTreeIndex
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from shapely.geometry.base import BaseMultipartGeometry
from shapely.ops import unary_union
from shapely.wkt import dumps

from fractopo.general import (
    CLASS_COLUMN,
    CONNECTION_COLUMN,
    GEOMETRY_COLUMN,
    JOBLIB_CACHE,
    CC_branch,
    CE_branch,
    CI_branch,
    E_node,
    EE_branch,
    Error_branch,
    I_node,
    IE_branch,
    II_branch,
    X_node,
    Y_node,
    check_for_z_coordinates,
    crop_to_target_areas,
    geom_bounds,
    get_trace_coord_points,
    get_trace_endpoints,
    line_intersection_to_points,
    numpy_to_python_type,
    point_to_point_unit_vector,
    pygeos_spatial_index,
    remove_z_coordinates_from_geodata,
    spatial_index_intersection,
)

log = logging.getLogger(__name__)

UNARY_ERROR_SIZE_THRESHOLD = 100


def determine_branch_identity(
    number_of_i_nodes: int, number_of_xy_nodes: int, number_of_e_nodes: int
) -> str:
    """
    Determine the identity of a branch.

    Is based on the amount of I-, XY- and E-nodes and returns it as a string.

    E.g.

    >>> determine_branch_identity(2, 0, 0)
    'I - I'

    >>> determine_branch_identity(1, 1, 0)
    'C - I'

    >>> determine_branch_identity(1, 0, 1)
    'I - E'

    """
    branch_type = Error_branch
    if number_of_i_nodes + number_of_e_nodes + number_of_xy_nodes != 2:
        log.error("Did not find 2 EXYI-nodes that intersected branch endpoints.\n")
        branch_type = Error_branch
    elif number_of_i_nodes == 2:
        branch_type = II_branch
    elif number_of_xy_nodes == 2:
        branch_type = CC_branch
    elif number_of_e_nodes == 2:
        branch_type = EE_branch
    elif number_of_i_nodes == 1 and number_of_xy_nodes == 1:
        branch_type = CI_branch
    elif number_of_e_nodes == 1 and number_of_xy_nodes == 1:
        branch_type = CE_branch
    elif number_of_e_nodes == 1 and number_of_i_nodes == 1:
        branch_type = IE_branch
    else:
        log.error("Unknown error in determine_branch_identity")
        branch_type = EE_branch
    return branch_type


def get_branch_identities(
    branches: gpd.GeoSeries,
    nodes: gpd.GeoSeries,
    node_identities: list,
    snap_threshold: float,
) -> List[str]:
    """
    Determine the types of branches for a GeoSeries of branches.

    i.e. C-C, C-I or I-I, + (C-E, E-E, I-E)

    >>> branches = gpd.GeoSeries(
    ...     [
    ...         LineString([(1, 1), (2, 2)]),
    ...         LineString([(2, 2), (3, 3)]),
    ...         LineString([(3, 0), (2, 2)]),
    ...         LineString([(2, 2), (-2, 5)]),
    ...     ]
    ... )
    >>> nodes = gpd.GeoSeries(
    ...     [
    ...         Point(2, 2),
    ...         Point(1, 1),
    ...         Point(3, 3),
    ...         Point(3, 0),
    ...         Point(-2, 5),
    ...     ]
    ... )
    >>> node_identities = ["X", "I", "I", "I", "E"]
    >>> snap_threshold = 0.001
    >>> get_branch_identities(branches, nodes, node_identities, snap_threshold)
    ['C - I', 'C - I', 'C - I', 'C - E']

    """
    assert len(nodes) == len(node_identities)
    node_spatial_index = pygeos_spatial_index(nodes)
    branch_identities = []
    for branch in branches.geometry.values:
        assert isinstance(branch, LineString)
        node_candidate_idxs = spatial_index_intersection(
            spatial_index=node_spatial_index, coordinates=geom_bounds(branch)
        )
        # node_candidate_idxs = list(node_spatial_index.intersection(branch.bounds))
        node_candidates = nodes.iloc[node_candidate_idxs]
        node_candidate_types = [node_identities[i] for i in node_candidate_idxs]

        # Use distance instead of two polygon buffers
        inter = [
            dist < snap_threshold
            for dist in node_candidates.distance(
                MultiPoint(list(get_trace_endpoints(branch)))
            ).values
        ]
        assert len(inter) == len(node_candidates)
        # nodes_that_intersect = node_candidates.loc[inter]
        nodes_that_intersect_types = list(compress(node_candidate_types, inter))
        number_of_E_nodes = sum(
            inter_id == E_node for inter_id in nodes_that_intersect_types
        )
        number_of_I_nodes = sum(
            inter_id == I_node for inter_id in nodes_that_intersect_types
        )
        number_of_XY_nodes = sum(
            inter_id in [X_node, Y_node] for inter_id in nodes_that_intersect_types
        )
        branch_identities.append(
            determine_branch_identity(
                number_of_I_nodes, number_of_XY_nodes, number_of_E_nodes
            )
        )

    return branch_identities


def angle_to_point(
    point: Point, nearest_point: Point, comparison_point: Point
) -> float:
    """
    Calculate the angle between two vectors.

    Vectors are made from the given points: Both vectors have the same first
    point, nearest_point, and second point is either point or comparison_point.

    Returns angle in degrees.

    E.g.

    >>> point = Point(1, 1)
    >>> nearest_point = Point(0, 0)
    >>> comparison_point = Point(-1, 1)
    >>> angle_to_point(point, nearest_point, comparison_point)
    90.0

    >>> point = Point(1, 1)
    >>> nearest_point = Point(0, 0)
    >>> comparison_point = Point(-1, 2)
    >>> angle_to_point(point, nearest_point, comparison_point)
    71.56505117707799

    """
    if (
        point.intersects(nearest_point)
        or point.intersects(comparison_point)
        or nearest_point.intersects(comparison_point)
    ):
        raise ValueError("Expected points in angle_to_point not to intersect.")

    unit_vector_1 = point_to_point_unit_vector(point=nearest_point, other_point=point)
    unit_vector_2 = point_to_point_unit_vector(
        point=nearest_point, other_point=comparison_point
    )
    dot_product = np.dot(unit_vector_1, unit_vector_2)
    if dot_product > 1 or dot_product < -1:
        rad_angle = np.nan
    else:
        rad_angle = np.arccos(dot_product)
    if np.isnan(rad_angle):
        # Cannot determine with angle.
        unit_vector_sum_len = np.linalg.norm(unit_vector_1 + unit_vector_2)
        if np.isclose(unit_vector_sum_len, 0.0, atol=1e-07):
            return 180.0
        if np.isclose(unit_vector_sum_len, 2.0, atol=1e-07):
            return 0.0
        vectors_dict = dict(
            unit_vector_1=unit_vector_1,
            unit_vector_2=unit_vector_2,
            unit_vector_sum_len=unit_vector_sum_len,
        )
        raise ValueError(
            f"Could not determine point relationships. Vectors: {vectors_dict}"
        )
    degrees = numpy_to_python_type(np.rad2deg(rad_angle))
    assert 360.0 >= degrees >= 0.0
    assert isinstance(degrees, float)
    return degrees


def insert_point_to_linestring(
    trace: LineString, point: Point, snap_threshold: float
) -> LineString:
    """
    Insert/modify point to trace LineString.

    The point location is determined to fit into the LineString without
    changing the geometrical order of LineString vertices
    (which only makes sense if LineString is sublinear.)

    TODO: Does not work for 2.5D geometries (Z-coordinates).
    Z-coordinates will be lost.

    E.g.

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> point = Point(1.25, 0.1)
    >>> insert_point_to_linestring(trace, point, 0.01).wkt
    'LINESTRING (0 0, 1 0, 1.25 0.1, 2 0, 3 0)'

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> point = Point(2.25, 0.1)
    >>> insert_point_to_linestring(trace, point, 0.01).wkt
    'LINESTRING (0 0, 1 0, 2 0, 2.25 0.1, 3 0)'

    """
    assert isinstance(trace, LineString)
    assert isinstance(point, Point)
    if trace.has_z:
        log.warning("Trace contains z-coordinates. These will be lost.")
    if point.has_z:
        log.warning("Point contains z-coordinates. These will be lost.")

    if any(point.intersects(Point(xy)) for xy in list(trace.coords)):
        log.error(
            "Point already matches a coordinate point in trace.\n"
            f"point: {point.wkt}, trace: {trace.wkt}\n"
            "Returning original trace without changes."
        )
        return trace
    trace_point_dists = [
        (idx, trace_point, trace_point.distance(point))
        for idx, trace_point in enumerate([Point(c) for c in list(trace.coords)])
    ]
    trace_point_dists = sorted(trace_point_dists, key=lambda vals: vals[2])
    nearest_point = trace_point_dists[0][1]
    nearest_point_idx = trace_point_dists[0][0]

    # Determine if to insert or modify trace
    idx, insert = determine_insert_approach(
        nearest_point_idx, trace_point_dists, snap_threshold, point, nearest_point
    )

    t_coords = list(trace.coords)
    if not insert:
        t_coords.pop(idx)
    t_coords.insert(idx, (point.x, point.y)) if not point.has_z else t_coords.insert(
        idx, (point.x, point.y, point.z)
    )
    # Closest points might not actually be the points which in between the
    # point is added. Have to use project and interpolate (?)
    # print(t_coords)
    new_trace = LineString(t_coords)
    # print(new_trace.wkt)
    assert new_trace.intersects(point)
    assert isinstance(new_trace, LineString)
    assert new_trace.is_valid
    # assert new_trace.is_simple
    if not new_trace.is_simple:
        log.warning(f"Non-simple geometry detected.\n{new_trace.wkt}")
    return new_trace


def determine_insert_approach(
    nearest_point_idx: int,
    trace_point_dists: List[Tuple[int, Point, float]],
    snap_threshold: float,
    point: Point,
    nearest_point: Point,
):
    """
    Determine if to insert or replace point.
    """
    insert = True
    if nearest_point_idx == 0:
        # It is the first node of linestring
        idx = nearest_point_idx + 1
    elif nearest_point_idx == max(vals[0] for vals in trace_point_dists):
        # It is the last node of linestring
        idx = nearest_point_idx
    else:
        nearest_point_distance_to_point = nearest_point.distance(point)
        if (
            nearest_point_distance_to_point < snap_threshold
            or nearest_point.wkt == point.wkt
            or np.isnan(nearest_point_distance_to_point)
        ):
            # Replace instead of insert
            insert = False
            idx = nearest_point_idx
        else:
            # It is in the middle of the linestring
            points_on_either_side = [
                (
                    vals,
                    angle_to_point(
                        point=point,
                        nearest_point=nearest_point,
                        comparison_point=vals[1],
                    ),
                )
                for vals in trace_point_dists
                if vals[0] in (nearest_point_idx - 1, nearest_point_idx + 1)
            ]

            points_on_either_side = sorted(
                points_on_either_side, key=lambda vals: vals[1]
            )

            assert sum(vals[1] == 0.0 for vals in points_on_either_side) < 2
            smallest_angle_idx = points_on_either_side[0][0][0]

            if smallest_angle_idx > nearest_point_idx:
                idx = smallest_angle_idx
            else:
                idx = nearest_point_idx
    return idx, insert


def additional_snapping_func(
    trace: LineString, idx: int, additional_snapping: List[Tuple[int, Point]]
) -> LineString:
    """
    Insert points into LineStrings to make sure trace abutting trace.

    E.g.

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> idx = 0
    >>> point = Point(2.25, 0.1)
    >>> additional_snapping = [
    ...     (0, point),
    ... ]
    >>> additional_snapping_func(trace, idx, additional_snapping).wkt
    'LINESTRING (0 0, 1 0, 2 0, 2.25 0.1, 3 0)'

    When idx doesn't match -> no additional snapping

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> idx = 1
    >>> point = Point(2.25, 0.1)
    >>> additional_snapping = [
    ...     (0, point),
    ... ]
    >>> additional_snapping_func(trace, idx, additional_snapping).wkt
    'LINESTRING (0 0, 1 0, 2 0, 3 0)'

    """
    indexes_to_fix, points_to_add = zip(*additional_snapping)
    if idx not in indexes_to_fix:
        return trace

    is_idx = np.array(indexes_to_fix) == idx
    filtered_points = compress(points_to_add, is_idx)
    for p in filtered_points:
        trace = insert_point_to_linestring(trace, p, snap_threshold=0.001)
    assert isinstance(trace, LineString)
    return trace


def snap_traces(
    traces: List[LineString],
    snap_threshold: float,
    areas: Optional[List[Union[Polygon, MultiPolygon]]] = None,
    final_allowed_loop=False,
) -> Tuple[List[LineString], bool]:
    """
    Snap traces to end exactly at other traces.
    """
    if len(traces) == 0:
        return ([], False)
    # Only handle LineStrings
    assert all(isinstance(trace, LineString) for trace in traces)

    # Spatial index for traces
    traces_spatial_index = pygeos_spatial_index(geodataset=gpd.GeoSeries(traces))

    # Collect simply snapped (and non-snapped) traces to list
    simply_snapped_traces, simple_changes = zip(
        *[
            snap_trace_simple(
                idx,
                trace,
                snap_threshold,
                traces,
                traces_spatial_index,
                final_allowed_loop=final_allowed_loop,
            )
            for idx, trace in enumerate(traces)
        ]
    )
    assert len(simply_snapped_traces) == len(traces)
    simply_snapped_traces_list: List[LineString] = list(simply_snapped_traces)

    # Collect snapped (and non-snapped) traces to list
    snapped_traces, changes = zip(
        *[
            snap_others_to_trace(
                idx=idx,
                trace=trace,
                snap_threshold=snap_threshold,
                traces_spatial_index=traces_spatial_index,
                areas=areas,
                traces=simply_snapped_traces_list,
                final_allowed_loop=final_allowed_loop,
            )
            for idx, trace in enumerate(simply_snapped_traces)
        ]
    )

    assert len(snapped_traces) == len(simply_snapped_traces)
    assert all(isinstance(ls, LineString) for ls in snapped_traces)

    return list(snapped_traces), any(changes + simple_changes)


def resolve_trace_candidates(
    trace: LineString,
    idx: int,
    traces_spatial_index: PyGEOSSTRTreeIndex,
    traces: List[LineString],
    snap_threshold: float,
) -> List[LineString]:
    """
    Resolve PyGEOSSTRTreeIndex intersection to actual intersection candidates.
    """
    assert isinstance(trace, LineString)

    # Get trace bounds and extend them
    # minx, miny, maxx, maxy = trace.bounds
    minx, miny, maxx, maxy = geom_bounds(trace)
    extended_bounds = (
        minx - snap_threshold * 20,
        miny - snap_threshold * 20,
        maxx + snap_threshold * 20,
        maxy + snap_threshold * 20,
    )

    # Use extended trace bounds to catch all possible intersecting traces
    trace_candidate_idxs_raw = list(traces_spatial_index.intersection(extended_bounds))
    trace_candidate_idxs = [
        i.item() for i in trace_candidate_idxs_raw if isinstance(i.item(), int)
    ]
    assert len(trace_candidate_idxs_raw) == len(trace_candidate_idxs)
    assert isinstance(trace_candidate_idxs, list)

    # Remove current trace
    trace_candidate_idxs.remove(idx)

    # Filter to only candidates based on spatial index
    # trace_candidates = traces.iloc[trace_candidate_idxs]
    # using operator.itemgetter() to
    # elements from list
    assert isinstance(trace_candidate_idxs, list)
    assert isinstance(traces, list)
    # trace_candidates = (
    #     list(itemgetter(*trace_candidate_idxs)(traces))
    #     if len(trace_candidate_idxs) > 0
    #     else []
    # )

    trace_candidates = [traces[i] for i in trace_candidate_idxs]

    return trace_candidates


def snap_trace_simple(
    idx: int,
    trace: LineString,
    snap_threshold: float,
    traces: List[LineString],
    traces_spatial_index: PyGEOSSTRTreeIndex,
    final_allowed_loop: bool = False,
) -> Tuple[LineString, bool]:
    """
    Determine whether and how to perform simple snap.
    """
    trace_candidates = resolve_trace_candidates(
        trace=trace,
        idx=idx,
        traces=traces,
        traces_spatial_index=traces_spatial_index,
        snap_threshold=snap_threshold,
    )

    # If no candidates -> no intersecting -> trace is isolated
    if len(trace_candidates) == 0:
        return trace, False

    # Handle simple case where trace can be extended to meet an absolute
    # coordinate point already in one of the candidates
    trace, was_simple_snapped = simple_snap(trace, trace_candidates, snap_threshold)

    # Debugging
    if final_allowed_loop and was_simple_snapped:
        log.error("In final_allowed_loop and still snapping:")
        log.error(f"{traces, trace}")

    # Return trace and information if it was changed
    return trace, was_simple_snapped


def snap_others_to_trace(
    idx: int,
    trace: LineString,
    snap_threshold: float,
    traces: List[LineString],
    traces_spatial_index: PyGEOSSTRTreeIndex,
    areas: Optional[List[Union[Polygon, MultiPolygon]]],
    final_allowed_loop: bool = False,
) -> Tuple[LineString, bool]:
    """
    Determine whether and how to snap `trace` to `traces`.

    E.g.

    Trace gets new coordinates to snap other traces to it:

    >>> idx = 0
    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> snap_threshold = 0.001
    >>> traces = [trace, LineString([(1.5, 3), (1.5, 0.00001)])]
    >>> traces_spatial_index = pygeos_spatial_index(gpd.GeoSeries(traces))
    >>> areas = None
    >>> snapped = snap_others_to_trace(
    ...     idx, trace, snap_threshold, traces, traces_spatial_index, areas
    ... )
    >>> snapped[0].wkt, snapped[1]
    ('LINESTRING (0 0, 1 0, 1.5 1e-05, 2 0, 3 0)', True)

    Trace itself is not snapped by snap_others_to_trace:

    >>> idx = 0
    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> snap_threshold = 0.001
    >>> traces = [trace, LineString([(3.0001, -3), (3.0001, 0), (3, 3)])]
    >>> traces_spatial_index = pygeos_spatial_index(gpd.GeoSeries(traces))
    >>> areas = None
    >>> snapped = snap_others_to_trace(
    ...     idx, trace, snap_threshold, traces, traces_spatial_index, areas
    ... )
    >>> snapped[0].wkt, snapped[1]
    ('LINESTRING (0 0, 1 0, 2 0, 3 0)', False)
    """
    trace_candidates = resolve_trace_candidates(
        trace=trace,
        idx=idx,
        traces=traces,
        traces_spatial_index=traces_spatial_index,
        snap_threshold=snap_threshold,
    )

    if trace in list(trace_candidates):
        error = "Found trace in trace_candidates."
        log.error(
            error,
            extra=dict(trace_wkt=trace.wkt, trace_candidates_len=len(trace_candidates)),
        )
        raise ValueError(error)

    # If no candidates -> no intersecting -> trace is isolated
    if len(trace_candidates) == 0:
        return trace, False

    # Get all endpoints of trace_candidates
    endpoints: List[Point] = list(
        chain(
            *[
                list(get_trace_endpoints(trace_candidate))
                for trace_candidate in trace_candidates
            ]
        )
    )

    # Filter endpoints out that are near to the area boundary
    if areas is not None:
        endpoints = [
            ep
            for ep in endpoints
            if not is_endpoint_close_to_boundary(
                ep, areas, snap_threshold=snap_threshold
            )
        ]

    # Add/replace endpoints into trace if they are within snap_threshold
    trace, was_snapped = snap_trace_to_another(
        trace_endpoints=endpoints, another=trace, snap_threshold=snap_threshold
    )

    # Debugging
    if final_allowed_loop and was_snapped:
        log.error("In final_allowed_loop and still snapping:")
        log.error(f"{traces, endpoints}")

    # Return trace and information if it was changed
    return trace, was_snapped


def simple_snap(
    trace: LineString, trace_candidates: List[LineString], snap_threshold: float
) -> Tuple[LineString, bool]:
    """
    Modify conditionally trace to snap to any of trace_candidates.

    E.g.

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3, 0)])
    >>> trace_candidates = gpd.GeoSeries(
    ...     [LineString([(3.0001, -3), (3.0001, 0), (3, 3)])]
    ... )
    >>> snap_threshold = 0.001
    >>> snapped = simple_snap(trace, trace_candidates, snap_threshold)
    >>> snapped[0].wkt, snapped[1]
    ('LINESTRING (0 0, 1 0, 2 0, 3.0001 0)', True)

    Do not snap overlapping.

    >>> trace = LineString([(0, 0), (1, 0), (2, 0), (3.0002, 0)])
    >>> trace_candidates = gpd.GeoSeries(
    ...     [LineString([(3.0001, -3), (3.0001, 0), (3, 3)])]
    ... )
    >>> snap_threshold = 0.001
    >>> snapped = simple_snap(trace, trace_candidates, snap_threshold)
    >>> snapped[0].wkt, snapped[1]
    ('LINESTRING (0 0, 1 0, 2 0, 3.0002 0)', False)
    """
    trace_endpoints = get_trace_endpoints(trace)
    traces_to_snap_to = [
        candidate
        for candidate in trace_candidates
        if any(
            endpoint.distance(candidate) < snap_threshold
            for endpoint in trace_endpoints
        )
    ]
    replace_endpoint: Dict[str, Point] = dict()
    for trace_to_snap_to in traces_to_snap_to:
        assert isinstance(trace_to_snap_to, LineString)

        # Get coordinate points of trace_to_snap_to expect endpoints
        coord_points = get_trace_coord_points(trace_to_snap_to)[1:-1]
        if len(coord_points) == 0:
            # Cannot snap trace to any point as there are only endpoints
            continue

        # Check both trace endpoints
        for endpoint in trace_endpoints:
            # Check for already existing snap
            if any(endpoint.intersects(coord_point) for coord_point in coord_points):
                # Already snapped
                continue

            # Get distances between endpoint in coord_points
            distances: List[float] = [
                coord_point.distance(endpoint) for coord_point in coord_points
            ]
            # If not within snap_threshold -> continue
            if not min(distances) < snap_threshold:
                continue

            # Get intersection points
            intersection_points = line_intersection_to_points(
                first=trace, second=trace_to_snap_to
            )
            # Check for overlapping snap
            if any(
                intersection_point.distance(endpoint) < snap_threshold
                for intersection_point in intersection_points
            ):
                # Overlapping snap
                # These are not fixed here and fallback to the other snapping
                # method
                continue

            sorted_points = sorted(
                zip(coord_points, distances), key=lambda vals: vals[1]
            )
            if endpoint.wkt in replace_endpoint:
                error = "Found endpoint in replace_endpoint dict."
                log.error(
                    error,
                    extra=(
                        dict(
                            endpoint_wkt=endpoint.wkt,
                            replace_endpoint_len=len(replace_endpoint),
                        )
                    ),
                )
                raise ValueError(error)
            assert endpoint.wkt not in replace_endpoint
            replace_endpoint[endpoint.wkt] = sorted_points[0][0]

    # If not replacing is done -> return original
    if len(replace_endpoint) == 0:
        return trace, False

    # Replace one of the endpoints based on replace_endpoint
    trace_coords = get_trace_coord_points(trace)
    trace_coords = [
        point if point.wkt not in replace_endpoint else replace_endpoint[point.wkt]
        for point in trace_coords
    ]

    # Return modified
    modified = LineString(trace_coords)
    return modified, True


def filter_non_unique_traces(
    traces: gpd.GeoSeries, snap_threshold: float
) -> gpd.GeoSeries:
    """
    Filter out traces that are not unique.
    """
    assert isinstance(traces, gpd.GeoSeries)
    traces_set = set()
    idxs_to_keep = []
    for idx, geom in enumerate(traces.geometry.values):
        assert isinstance(geom, LineString), geom

        # Use a less strict coordinate precision when finding duplicates
        geom_wkt = dumps(geom, rounding_precision=int(-math.log10(snap_threshold)))
        if geom_wkt in traces_set:
            continue
        traces_set.add(geom_wkt)
        idxs_to_keep.append(idx)

    unique_traces = traces.iloc[idxs_to_keep]
    assert isinstance(unique_traces, gpd.GeoSeries)

    filter_count = len(traces) - len(unique_traces)
    log.info(
        f"Filtered out {filter_count} traces.",
        extra=dict(traces_len=len(traces), unique_traces_len=len(unique_traces)),
    )
    return unique_traces


def safer_unary_union(
    traces_geosrs: gpd.GeoSeries, snap_threshold: float, size_threshold: int
) -> MultiLineString:
    """
    Perform unary union to transform traces to branch segments.

    unary_union is not completely stable with large datasets but problem can be
    alleviated by dividing analysis to parts.

    TODO: Usage is deprecated as unary_union seems to give consistent results.
    """
    if traces_geosrs.empty:
        return MultiLineString()

    # Get amount of traces
    trace_count = traces_geosrs.shape[0]

    # Only one trace and self-intersects shouldn't occur -> Simply return the
    # one LineString wrapped in MultiLineString
    if trace_count == 1:
        return MultiLineString(list(traces_geosrs.geometry.values))

    # Try normal union without any funny business
    # This will be compared to the split approach result and better will
    # be returned
    normal_full_union: MultiLineString = traces_geosrs.unary_union

    if isinstance(normal_full_union, LineString):
        return MultiLineString([normal_full_union])

    if trace_count < size_threshold:
        if len(normal_full_union.geoms) > trace_count and isinstance(
            normal_full_union, MultiLineString
        ):
            return normal_full_union

    # Debugging, fail safely
    if size_threshold < UNARY_ERROR_SIZE_THRESHOLD:
        log.critical(
            "Expected size_threshold to be higher than 100. Union might be impossible."
        )

    # How many parts
    div = int(np.ceil(trace_count / size_threshold))

    # Divide with numpy
    split_traces = np.array_split(traces_geosrs, div)
    assert isinstance(split_traces, list)

    # How many in each pair
    # part_count = int(np.ceil(trace_count / div))

    assert div * sum(part.shape[0] for part in split_traces) >= trace_count
    assert all(isinstance(val, gpd.GeoSeries) for val in split_traces)
    assert isinstance(split_traces[0].iloc[0], LineString)

    # Do unary_union in parts
    part_unions = part_unary_union(
        split_traces=split_traces,
        snap_threshold=snap_threshold,
        size_threshold=size_threshold,
        div=div,
    )
    # Do full union of split unions
    full_union = unary_union(MultiLineString(list(chain(*part_unions))))

    # full_union should always be better or equivalent to normal unary_union.
    # (better when unary_union fails silently)
    if isinstance(full_union, MultiLineString):
        assert isinstance(full_union, BaseMultipartGeometry)
        if len(full_union.geoms) >= len(normal_full_union.geoms):
            return full_union

        raise ValueError(
            "Expected split union to give better results."
            " Branches and nodes should be checked for inconsistencies."
        )
    if isinstance(full_union, LineString):
        return MultiLineString([full_union])
    raise TypeError(
        f"Expected (Multi)LineString from unary_union. Got {full_union.wkt}"
    )


def part_unary_union(
    split_traces: list, snap_threshold: float, size_threshold: int, div: int
):
    """
    Conduct safer_unary_union in parts.
    """
    # Collect partly done unary_unions to part_unions list
    part_unions = []

    # Iterate over list of split trace GeoSeries
    for part in split_traces:
        # Do unary_union to part
        part_union = part.unary_union

        # Do naive check for if unary_union is successful
        if (
            not isinstance(part_union, MultiLineString)
            or len(part_union.geoms) < part.shape[0]
        ):
            # Still fails -> Try with lower threshold for part
            part_union = safer_unary_union(
                part, snap_threshold, size_threshold=size_threshold // 2
            )

        # Collect
        part_unions.append(part_union.geoms)
    assert len(part_unions) == div
    return part_unions


def report_snapping_loop(loops: int, allowed_loops: int):
    """
    Report snapping looping.
    """
    log.info(f"Loop :{ loops }")
    if loops >= 10:
        log.warning(
            f"{loops} loops have passed without resolved snapped"
            " traces_geosrs. Snapped traces_geosrs might not"
            " possibly be resolved."
        )
    if loops > allowed_loops:
        raise RecursionError(
            f"More loops have passed ({loops}) than allowed by allowed_loops "
            f"({allowed_loops})) for snapping traces_geosrs for"
            " branch determination."
        )


@JOBLIB_CACHE.cache
def branches_and_nodes(
    traces: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    areas: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    snap_threshold: float,
    allowed_loops=10,
    already_clipped: bool = False,
    # unary_size_threshold: int = 5000,
) -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """
    Determine branches and nodes of given traces.

    The traces will be cropped to the given target area(s) if not already
    clipped(already_clipped).

    TODO: unary_union will not respect identical traces or near-identical.
    Therefore cannot test if there are more branches than traces because
    there might be less due to this issue.
    """
    log.info(
        "Starting determination of branches and nodes.",
        extra=dict(
            len_traces=len(traces),
            len_areas=len(areas),
            snap_threshold=snap_threshold,
            already_clipped=already_clipped,
            allowed_loops=allowed_loops,
        ),
    )
    traces_geosrs: gpd.GeoSeries = traces.geometry
    if check_for_z_coordinates(geodata=traces_geosrs):
        log.info(
            "Removing z-coordinates from trace data before branch and node determination."
        )
        traces_without_z_coords = remove_z_coordinates_from_geodata(
            geodata=traces_geosrs
        )
        assert isinstance(traces_without_z_coords, gpd.GeoSeries)
        traces_geosrs = traces_without_z_coords

    # Filter out traces that are not unique by wkt
    # unary_union will fail to take them into account any way
    traces_geosrs = filter_non_unique_traces(
        traces_geosrs, snap_threshold=snap_threshold
    )

    areas_geosrs: gpd.GeoSeries = areas.geometry

    # Collect into lists
    areas_lists_of_polygons = [
        [poly] if isinstance(poly, Polygon) else list(poly.geoms)
        for poly in areas_geosrs.geometry.values
        if isinstance(poly, (Polygon, MultiPolygon))
    ]
    areas_list = list(chain(*areas_lists_of_polygons))
    assert all(isinstance(poly, Polygon) for poly in areas_list), areas_list

    # Collect into lists
    traces_list = [
        trace
        for trace in traces_geosrs.geometry.values
        if isinstance(trace, LineString)
    ]

    # Snapping occurs multiple times due to possible side effects of each snap
    loops = 0
    traces_list, any_changes_applied = snap_traces(
        traces_list, snap_threshold, areas=areas_list
    )
    # Snapping causes changes that might cause new errors. The snapping is looped
    # as many times as there are changed made to the data.
    # If loop count reaches allowed_loops, error is raised.
    while any_changes_applied:
        traces_list, any_changes_applied = snap_traces(
            traces_list,
            snap_threshold,
            final_allowed_loop=loops == allowed_loops,
            areas=areas_list,
        )
        loops += 1
        report_snapping_loop(loops, allowed_loops=allowed_loops)

    traces_geosrs = gpd.GeoSeries(traces_list, crs=traces.crs)

    # Clip if necessary
    if not already_clipped:
        traces_geosrs = crop_to_target_areas(
            traces_geosrs,
            areas_geosrs,
            keep_column_data=False,
        ).geometry

    # Remove too small geometries.
    traces_geosrs = traces_geosrs.loc[
        traces_geosrs.geometry.length > snap_threshold * 2.01
    ]

    # Branches are determined with shapely/geopandas unary_union
    unary_union_result = traces_geosrs.unary_union
    if isinstance(unary_union_result, MultiLineString):
        branches_all = list(unary_union_result.geoms)
    elif isinstance(unary_union_result, LineString):
        branches_all = [unary_union_result]
    else:
        raise TypeError(
            "Expected unary_union_result to be of type (Multi)LineString."
            f" Got: {type(unary_union_result), unary_union_result}"
        )

    # branches_all = list(
    #     safer_unary_union(
    #         traces_geosrs,
    #         snap_threshold=snap_threshold,
    #         size_threshold=unary_size_threshold,
    #     ).geoms
    # )

    # Filter out very short branches
    branches = gpd.GeoSeries(
        [b for b in branches_all if b.length > snap_threshold * 1.01],
        crs=traces_geosrs.crs,
    )

    # Report and error possibly unary_union failure
    if len(branches_all) < len(traces_geosrs):
        # unary_union can fail with too large datasets
        log.critical(
            "Expected more branches than traces. Possible unary_union failure."
        )

    # Determine nodes and identities
    nodes, node_identities = node_identities_from_branches(
        branches=branches, areas=areas_geosrs, snap_threshold=snap_threshold
    )

    # Collect to GeoSeries
    nodes_geosrs = gpd.GeoSeries(nodes)

    # Determine branch identities
    branch_identities = get_branch_identities(
        branches, nodes_geosrs, node_identities, snap_threshold
    )

    # Collect to GeoDataFrames
    node_gdf = gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: nodes_geosrs, CLASS_COLUMN: node_identities}, crs=traces.crs
    )
    branch_gdf = gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: branches, CONNECTION_COLUMN: branch_identities},
        crs=traces.crs,
    )
    return branch_gdf, node_gdf


def snap_trace_to_another(
    trace_endpoints: List[Point], another: LineString, snap_threshold: float
) -> Tuple[LineString, bool]:
    """
    Add point to another trace to snap trace to end at another trace.

    I.e. modifies and returns `another`
    """
    # Do nothing if endpoints of trace are not within snap_threshold.
    # Or if endpoint is already exact
    endpoints = [
        ep
        for ep in trace_endpoints
        if ep.distance(another) < snap_threshold and not ep.intersects(another)
    ]

    # No endpoints within snapping threshold
    if len(endpoints) == 0:
        # Do not snap
        return another, False

    for endpoint in endpoints:
        another = insert_point_to_linestring(
            another, endpoint, snap_threshold=snap_threshold
        )

    return another, True


def is_endpoint_close_to_boundary(
    endpoint: Point, areas: List[Union[Polygon, MultiPolygon]], snap_threshold: float
) -> bool:
    """
    Check if endpoint is within snap_threshold of areas boundaries.
    """
    for area in areas:
        assert isinstance(area, (Polygon, MultiPolygon))
        if endpoint.distance(area.boundary) < snap_threshold:
            return True
    return False


def node_identity(
    endpoint: Point,
    idx: int,
    areas: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    endpoints_geoseries: gpd.GeoSeries,
    endpoints_spatial_index: PyGEOSSTRTreeIndex,
    snap_threshold: float,
) -> str:
    """
    Determine node identity of endpoint.
    """
    node_type = E_node
    # Check for proximity to area boundary
    if any(
        endpoint.distance(area.boundary) < snap_threshold
        for area in areas.geometry.values
    ):
        return node_type

    # Candidate points idxs
    candidate_idxs = list(endpoints_spatial_index.intersection(endpoint.coords[0]))

    # Remove current
    candidate_idxs.remove(idx)

    # Candidate points
    candidates = endpoints_geoseries.iloc[candidate_idxs]

    # Get intersect count
    intersecting_node_count = sum(
        candidate.distance(endpoint) < snap_threshold for candidate in candidates
    )

    if intersecting_node_count == 0:
        node_type = I_node
    elif intersecting_node_count == 2:
        node_type = Y_node
    elif intersecting_node_count == 3:
        node_type = X_node
    elif intersecting_node_count == 1:
        # Unresolvable
        log.error(
            f"Expected 0, 2 or 3 intersects. V-node or similar error at {endpoint.wkt}."
        )
        node_type = I_node
    else:
        # Unresolvable
        log.error(f"Expected 0, 2 or 3 intersects. Multijunction at {endpoint.wkt}.")
        node_type = X_node
    return node_type


def node_identities_from_branches(
    branches: gpd.GeoSeries,
    areas: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    snap_threshold: float,
) -> Tuple[List[Point], List[str]]:
    """
    Resolve node identities from branch data.

    >>> branches_list = [
    ...     LineString([(0, 0), (1, 1)]),
    ...     LineString([(2, 2), (1, 1)]),
    ...     LineString([(2, 0), (1, 1)]),
    ... ]
    >>> area_polygon = Polygon([(-5, -5), (-5, 5), (5, 5), (5, -5)])
    >>> branches = gpd.GeoSeries(branches_list)
    >>> areas = gpd.GeoSeries([area_polygon])
    >>> snap_threshold = 0.001
    >>> nodes, identities = node_identities_from_branches(
    ...     branches, areas, snap_threshold
    ... )
    >>> [node.wkt for node in nodes]
    ['POINT (0 0)', 'POINT (1 1)', 'POINT (2 2)', 'POINT (2 0)']
    >>> identities
    ['I', 'Y', 'I', 'I']

    """
    # Get list of all branch endpoints
    all_endpoints: List[Point] = list(
        chain(
            *[list(get_trace_endpoints(branch)) for branch in branches.geometry.values]
        )
    )
    # Collect into GeoSeries
    all_endpoints_geoseries = gpd.GeoSeries(all_endpoints)

    # Get spatial index
    endpoints_spatial_index = pygeos_spatial_index(all_endpoints_geoseries)

    # Collect resolved nodes
    collected_nodes: Dict[str, Tuple[Point, str]] = dict()

    for idx, endpoint in enumerate(all_endpoints):
        # Do not resolve nodes that have already been resolved
        if endpoint.wkt in collected_nodes:
            continue

        # Determine node identity
        identity = node_identity(
            endpoint=endpoint,
            idx=idx,
            areas=areas,
            endpoints_geoseries=all_endpoints_geoseries,
            endpoints_spatial_index=endpoints_spatial_index,
            snap_threshold=snap_threshold,
        )

        # Add to resolved
        collected_nodes[endpoint.wkt] = (endpoint, identity)

    if len(collected_nodes) == 0:
        return [], []

    # Collect into two lists
    values = list(collected_nodes.values())
    nodes = [value[0] for value in values]
    identities = [value[1] for value in values]

    return nodes, identities



"""
Scripts for creating sample grids for fracture trace, branch and node data.
"""
import logging
import platform
from typing import Dict, Optional

import geopandas as gpd
import numpy as np
from geopandas.sindex import PyGEOSSTRTreeIndex
from joblib import Parallel, delayed
from shapely.geometry import LineString, Point, Polygon

from fractopo.analysis.parameters import (
    determine_node_type_counts,
    determine_topology_parameters,
)
from fractopo.branches_and_nodes import branches_and_nodes
from fractopo.general import (
    CLASS_COLUMN,
    GEOMETRY_COLUMN,
    JOBLIB_CACHE,
    Param,
    crop_to_target_areas,
    geom_bounds,
    pygeos_spatial_index,
    safe_buffer,
    spatial_index_intersection,
)

log = logging.getLogger(__name__)


def create_grid(cell_width: float, lines: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Create an empty polygon grid for sampling fracture line data.

    Grid is created to always contain all given lines.

    E.g.

    >>> lines = gpd.GeoSeries(
    ...     [
    ...         LineString([(1, 1), (2, 2)]),
    ...         LineString([(2, 2), (3, 3)]),
    ...         LineString([(3, 0), (2, 2)]),
    ...         LineString([(2, 2), (-2, 5)]),
    ...     ]
    ... )
    >>> create_grid(cell_width=0.1, lines=lines).head(5)
                                                geometry
    0  POLYGON ((-2.00000 5.00000, -1.90000 5.00000, ...
    1  POLYGON ((-2.00000 4.90000, -1.90000 4.90000, ...
    2  POLYGON ((-2.00000 4.80000, -1.90000 4.80000, ...
    3  POLYGON ((-2.00000 4.70000, -1.90000 4.70000, ...
    4  POLYGON ((-2.00000 4.60000, -1.90000 4.60000, ...
    """
    assert cell_width > 0
    assert len(lines) > 0
    assert all(isinstance(val, LineString) for val in lines.geometry.values)

    # Get total bounds of branches
    x_min, y_min, x_max, y_max = lines.total_bounds
    cell_height = cell_width

    # Calculate cell row and column counts
    rows = int(np.ceil((y_max - y_min) / cell_height))
    cols = int(np.ceil((x_max - x_min) / cell_width))

    x_left_origin = x_min
    x_right_origin = x_min + cell_width
    y_top_origin = y_max
    y_bottom_origin = y_max - cell_height
    polygons = []
    # Create grid cell polygons
    for _ in range(cols):
        y_top = y_top_origin
        y_bottom = y_bottom_origin
        for _ in range(rows):
            polygons.append(
                Polygon(
                    [
                        (x_left_origin, y_top),
                        (x_right_origin, y_top),
                        (x_right_origin, y_bottom),
                        (x_left_origin, y_bottom),
                    ]
                )
            )
            y_top = y_top - cell_height
            y_bottom = y_bottom - cell_height
        x_left_origin = x_left_origin + cell_width
        x_right_origin = x_right_origin + cell_width

    # Create GeoDataFrame with grid polygons
    grid = gpd.GeoDataFrame({GEOMETRY_COLUMN: polygons}, crs=lines.crs)
    assert len(grid) != 0
    return grid


def populate_sample_cell(
    sample_cell: Polygon,
    sample_cell_area: float,
    traces: gpd.GeoDataFrame,
    nodes: gpd.GeoDataFrame,
    branches: gpd.GeoDataFrame,
    snap_threshold: float,
    resolve_branches_and_nodes: bool,
    traces_sindex: Optional[PyGEOSSTRTreeIndex] = None,
) -> Dict[str, float]:
    """
    Take a single grid polygon and populate it with parameters.

    Mauldon determination requires that E-nodes are defined for
    every single sample circle. If correct Mauldon values are
    wanted `resolve_branches_and_nodes` must be passed as True.
    This will result in much longer analysis time.
    """

    def resolve_samples(candidates, sample_circle):
        """
        Crop traces or branches to sample circle

        First check if any geometries intersect.
        If not: sample_features is an empty GeoDataFrame.
        """
        if any(
            candidate.intersects(sample_circle)
            for candidate in candidates.geometry.values
        ):
            samples = crop_to_target_areas(
                traces=candidates,
                areas=gpd.GeoSeries([sample_circle]),
                is_filtered=True,
                keep_column_data=False,
            )
        else:
            samples = candidates.iloc[0:0]
        return samples

    def choose_geometries(sindex, sample_circle, geometries):
        candidates_idx = spatial_index_intersection(sindex, geom_bounds(sample_circle))
        candidates = geometries.iloc[candidates_idx]
        assert isinstance(candidates, gpd.GeoDataFrame)
        return candidates

    centroid = sample_cell.centroid
    if not isinstance(centroid, Point):
        raise TypeError("Expected Point centroid.")
    sample_circle = safe_buffer(centroid, np.sqrt(sample_cell_area) * 1.5)
    sample_circle_area = sample_circle.area
    assert sample_circle_area > 0

    if traces_sindex is None:
        traces_sindex = pygeos_spatial_index(traces)

    # Choose geometries that are either within the sample_circle or
    # intersect it
    # Use spatial indexing to filter to only spatially relevant traces,
    # traces and nodes
    # trace_candidates_idx = spatial_index_intersection(
    #     traces_sindex, geom_bounds(sample_circle)
    # )
    trace_candidates = choose_geometries(
        sindex=traces_sindex, sample_circle=sample_circle, geometries=traces
    )

    if len(trace_candidates) == 0:
        return determine_topology_parameters(
            trace_length_array=np.array([]),
            node_counts=determine_node_type_counts(np.array([]), branches_defined=True),
            area=sample_circle_area,
            branch_length_array=np.array([]),
            branches_defined=True,
            correct_mauldon=resolve_branches_and_nodes,
        )
    if resolve_branches_and_nodes:
        # Solve branches and nodes for each cell if wanted
        # Only way to make sure Mauldon parameters are correct
        branches, nodes = branches_and_nodes(
            traces=trace_candidates,
            areas=gpd.GeoSeries([sample_circle], crs=traces.crs),
            snap_threshold=snap_threshold,
        )

    is_topology_defined = branches.shape[0] > 0
    if is_topology_defined:
        branch_candidates = choose_geometries(
            sindex=pygeos_spatial_index(branches),
            sample_circle=sample_circle,
            geometries=branches,
        )

        node_candidates = choose_geometries(
            sindex=pygeos_spatial_index(nodes),
            sample_circle=sample_circle,
            geometries=nodes,
        )
        sample_branches = resolve_samples(
            candidates=branch_candidates, sample_circle=sample_circle
        )
        if any(node.intersects(sample_circle) for node in nodes.geometry.values):
            # if any(nodes.intersects(sample_circle)):
            # TODO: Is node clipping stable?
            sample_nodes = gpd.clip(node_candidates, sample_circle)
            assert sample_nodes is not None
            assert all(isinstance(val, Point) for val in sample_nodes.geometry.values)
        else:
            sample_nodes = nodes.iloc[0:0]
        assert isinstance(sample_nodes, gpd.GeoDataFrame)
        assert isinstance(sample_branches, gpd.GeoDataFrame)
        sample_node_type_values = sample_nodes[CLASS_COLUMN].values
        assert isinstance(sample_node_type_values, np.ndarray)
        node_counts = determine_node_type_counts(
            sample_node_type_values, branches_defined=True
        )
        sample_branches_lengths = sample_branches.geometry.length.values
    else:
        sample_branches_lengths = np.array([])
        node_counts = determine_node_type_counts(np.array([]), branches_defined=True)

    sample_traces = resolve_samples(
        candidates=trace_candidates, sample_circle=sample_circle
    )

    assert isinstance(sample_traces, gpd.GeoDataFrame)

    topology_parameters = determine_topology_parameters(
        trace_length_array=sample_traces.geometry.length.values,
        branch_length_array=sample_branches_lengths,
        node_counts=node_counts,
        area=sample_circle_area,
        correct_mauldon=resolve_branches_and_nodes,
        # TODO: Sample grids are currently not implemented for Networks
        # without determined topology.
        branches_defined=is_topology_defined,
    )
    return topology_parameters


def sample_grid(
    grid: gpd.GeoDataFrame,
    traces: gpd.GeoDataFrame,
    nodes: gpd.GeoDataFrame,
    branches: gpd.GeoDataFrame,
    snap_threshold: float,
    resolve_branches_and_nodes: bool = False,
) -> gpd.GeoDataFrame:
    """
    Populate a sample polygon grid with geometrical and topological parameters.
    """
    sample_cell_area = grid.geometry.iloc[0].area
    assert sample_cell_area != 0
    # String identifiers as parameters
    # dict with lists of parameter values
    params = dict()  # type: Dict[str, list]
    # Iterate over sample cells
    # Uses a buffer 1.5 times the size of the sample_cell side length
    # Make sure index does not cause issues TODO
    traces_reset, nodes_reset, branches_reset = (
        traces.reset_index(drop=True),
        nodes.reset_index(drop=True),
        branches.reset_index(drop=True),
    )
    assert isinstance(traces_reset, gpd.GeoDataFrame)
    assert isinstance(nodes_reset, gpd.GeoDataFrame)
    assert isinstance(branches_reset, gpd.GeoDataFrame)
    traces, nodes, branches = traces_reset, nodes_reset, branches_reset
    # [gdf.reset_index(inplace=True, drop=True) for gdf in (traces, nodes)]
    # traces_sindex = pygeos_spatial_index(traces)
    # nodes_sindex = pygeos_spatial_index(nodes)

    # params_for_cells = list(
    #     map(
    #         lambda sample_cell: populate_sample_cell(
    #             sample_cell=sample_cell,
    #             sample_cell_area=sample_cell_area,
    #             traces_sindex=traces_sindex,
    #             traces=traces,
    #             nodes=nodes,
    #             branches=branches,
    #             snap_threshold=snap_threshold,
    #             resolve_branches_and_nodes=resolve_branches_and_nodes,
    #         ),
    #         grid.geometry.values,
    #     )
    # )
    # Use all CPUs with n_jobs=-1
    # Use only one process on Windows
    params_for_cells = Parallel(
        n_jobs=(-1 if not platform.system() == "Windows" else 1)
    )(
        delayed(populate_sample_cell)(
            sample_cell=sample_cell,
            sample_cell_area=sample_cell_area,
            traces_sindex=None,
            traces=traces,
            nodes=nodes,
            branches=branches,
            snap_threshold=snap_threshold,
            resolve_branches_and_nodes=resolve_branches_and_nodes,
        )
        for sample_cell in grid.geometry.values
    )
    assert isinstance(params_for_cells, list)
    for key in [param.value.name for param in Param]:
        assert isinstance(key, str)
        try:
            params[key] = [cell_param[key] for cell_param in params_for_cells]
        except KeyError:
            continue
    for key, value in params.items():
        grid[key] = value
    return grid


@JOBLIB_CACHE.cache
def run_grid_sampling(
    traces: gpd.GeoDataFrame,
    branches: gpd.GeoDataFrame,
    nodes: gpd.GeoDataFrame,
    cell_width: float,
    snap_threshold: float,
    precursor_grid: Optional[gpd.GeoDataFrame] = None,
    resolve_branches_and_nodes: bool = False,
) -> gpd.GeoDataFrame:
    """
    Run the contour grid sampling to passed trace, branch and node data.

    The grid extents are determined by using the passed branches unless
    precursor_grid is passed.

    If precursor_grid is passed, cell_width is not requires. Otherwise
    it must always be set to be to a non-default value.
    """
    if traces.empty:
        log.warning("Empty GeoDataFrame passed to run_grid_sampling.")
        return gpd.GeoDataFrame()
    if precursor_grid is not None:
        if not isinstance(precursor_grid, gpd.GeoDataFrame):
            raise TypeError("Expected precursor_grid to be of type: GeoDataFrame.")
        # Avoid modifying same precursor_grid multiple times
        grid = precursor_grid.copy()
    else:
        if np.isclose(cell_width, 0.0) or cell_width < 0:
            raise ValueError(
                "Expected cell_width to be non-close-to-zero positive number."
            )
        is_topology_defined = branches.shape[0] > 0
        if is_topology_defined:
            lines = branches
        else:
            lines = traces
        grid = create_grid(cell_width, lines=lines)
    sampled_grid = sample_grid(
        grid,
        traces,
        nodes,
        branches=branches,
        snap_threshold=snap_threshold,
        resolve_branches_and_nodes=resolve_branches_and_nodes,
    )
    return sampled_grid



"""
Utilities for analyzing and plotting length distributions for line data.
"""
import logging
from dataclasses import dataclass
from enum import Enum, unique
from itertools import chain, cycle
from textwrap import fill
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

import numpy as np
import powerlaw
import seaborn as sns
import sklearn.metrics as sklm
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.optimize import OptimizeResult, shgo
from sklearn.linear_model import LinearRegression

from fractopo import general

log = logging.getLogger(__name__)

ALPHA = "alpha"
EXPONENT = "exponent"
CUT_OFF = "cut-off"
KOLM_DIST = "Kolmogorov-Smirnov distance D"
SIGMA = "sigma"
MU = "mu"
LAMBDA = "lambda"
LOGLIKELIHOOD = "loglikelihood"

SCORER_NAMES: Dict[Callable, str] = {
    sklm.mean_squared_log_error: "MSLE",
    sklm.r2_score: "$R^2$",
    general.r2_scorer: "$R^2$ (inv.)",
}


class SilentFit(powerlaw.Fit):

    """
    Wrap powerlaw.Fit for the singular purpose of silencing output.

    Silences output both to stdout and stderr.
    """

    def __init__(
        self,
        data: np.ndarray,
        discrete=False,
        xmin=None,
        xmax=None,
        verbose=True,
        fit_method="Likelihood",
        estimate_discrete=True,
        discrete_approximation="round",
        sigma_threshold=None,
        parameter_range=None,
        fit_optimizer=None,
        xmin_distance="D",
        xmin_distribution="power_law",
        **kwargs,
    ):
        """
        Override Fit.__init__ to silence output.
        """
        with general.silent_output("__init__"):
            super().__init__(
                data,
                discrete=discrete,
                xmin=xmin,
                xmax=xmax,
                verbose=verbose,
                fit_method=fit_method,
                estimate_discrete=estimate_discrete,
                discrete_approximation=discrete_approximation,
                sigma_threshold=sigma_threshold,
                parameter_range=parameter_range,
                fit_optimizer=fit_optimizer,
                xmin_distance=xmin_distance,
                xmin_distribution=xmin_distribution,
                **kwargs,
            )

    def __getattribute__(self, name):
        """
        Get attribute with silent output.

        Also wraps all callables (~instance methods) with silent_output. The
        stdout and stderr is reported with log.info so it is not lost.
        """
        with general.silent_output("__getattribute__"):
            attribute = super().__getattribute__(name)
        if callable(attribute):
            return general.wrap_silence(attribute)
        return attribute


@unique
class Dist(Enum):

    """
    Enums of powerlaw model types.
    """

    POWERLAW = "power_law"
    LOGNORMAL = "lognormal"
    EXPONENTIAL = "exponential"
    TRUNCATED_POWERLAW = "truncated_power_law"


def numpy_polyfit(log_lengths: np.ndarray, log_ccm: np.ndarray) -> Tuple[float, float]:
    """
    Fit numpy polyfit to data.
    """
    vals = np.polyfit(log_lengths, log_ccm, 1)
    assert len(vals) == 2
    if isinstance(vals, tuple):
        return vals
    if isinstance(vals, np.ndarray):
        return vals[0], vals[1]
    raise TypeError(
        "Expected np.polyfit results to be a tuple or an array."
        f" Got {vals} with type: {type(vals)}."
    )


@dataclass
class LengthDistribution:

    """
    Dataclass for length distributions.
    """

    lengths: np.ndarray
    area_value: float
    using_branches: bool
    name: str = ""

    _automatic_fit: Optional[powerlaw.Fit] = None

    def __post_init__(self):
        """
        Filter lengths lower than general.MINIMUM_LINE_LENGTH.

        Also log the creation parameters.
        """
        assert self.area_value > 0
        # Filter lengths lower than general.MINIMUM_LINE_LENGTH.
        # Lengths lower than the value can cause runtime issues.
        filtered_lengths = self.lengths[self.lengths > general.MINIMUM_LINE_LENGTH]

        # Calculate proportion for log purposes
        filtered_proportion = (
            ((len(self.lengths) - len(filtered_lengths)) / len(self.lengths))
            if len(self.lengths) > 0
            else 0.0
        )
        high_filtered = filtered_proportion > 0.1
        log_func = log.warning if high_filtered else log.info
        log_func(
            "Created LengthDistribution instance."
            + (" High filter proportion!" if high_filtered else "")
        )

        # Set preprocessed lengths back into lengths attribute
        self.lengths = filtered_lengths

    @property
    def automatic_fit(self) -> Optional[powerlaw.Fit]:
        """
        Get automatic powerlaw Fit.
        """
        if self._automatic_fit is None:
            self._automatic_fit = determine_fit(length_array=self.lengths)
        return self._automatic_fit

    def manual_fit(self, cut_off: float):
        """
        Get manual powerlaw Fit.
        """
        return determine_fit(length_array=self.lengths, cut_off=cut_off)

    def generate_distributions(self, cut_off: float = general.MINIMUM_LINE_LENGTH):
        """
        Generate ccdf and truncated length data with cut_off.
        """
        lengths, ccm = sorted_lengths_and_ccm(
            lengths=self.lengths, area_value=self.area_value
        )
        lengths, ccm = apply_cut_off(lengths=lengths, ccm=ccm, cut_off=cut_off)
        return lengths, ccm


class Polyfit(NamedTuple):

    """
    Results of a polyfit to length data.
    """

    y_fit: np.ndarray
    m_value: float
    constant: float
    score: float
    scorer: Callable[[np.ndarray, np.ndarray], float]


class MultiScaleOptimizationResult(NamedTuple):

    """
    Results of scipy.optimize.shgo on length data.
    """

    polyfit: Polyfit
    cut_offs: np.ndarray
    optimize_result: OptimizeResult
    x0: np.ndarray
    bounds: np.ndarray
    proportions_of_data: List[float]


@dataclass
class MultiLengthDistribution:

    """
    Multi length distribution.
    """

    distributions: List[LengthDistribution]
    using_branches: bool
    fitter: Callable[[np.ndarray, np.ndarray], Tuple[float, float]] = numpy_polyfit
    cut_offs: Optional[List[float]] = None

    # Private caching attributes
    _fit_to_multi_scale_lengths: Optional[Tuple[np.ndarray, float, float]] = None
    _normalized_distributions: Optional[
        Tuple[List[np.ndarray], List[np.ndarray]]
    ] = None
    _optimized: bool = False

    def __hash__(self) -> int:
        """
        Implement hashing for MultiLengthDistribution.
        """
        all_lengths = tuple(chain(*[ld.lengths for ld in self.distributions]))
        all_lengths_str = tuple(map(str, all_lengths))
        all_area_values_str = tuple(map(str, [ld.name for ld in self.distributions]))
        cut_offs_tuple = (
            self.cut_offs if self.cut_offs is None else tuple(self.cut_offs)
        )
        return hash(
            (
                all_lengths_str,
                cut_offs_tuple,
                self.using_branches,
                all_area_values_str,
                self.fitter.__name__,
            )
        )

    def normalized_distributions(
        self, automatic_cut_offs: bool
    ) -> Tuple[
        List[np.ndarray],
        List[np.ndarray],
        List[np.ndarray],
        List[np.ndarray],
        List[np.ndarray],
    ]:
        """
        Create normalized and truncated lengths and ccms.
        """
        # Collect length and ccm arrays after their normalization
        (
            truncated_length_array_all,
            ccm_array_normed_all,
            full_length_array_all,
            full_ccm_array_normed_all,
            cut_offs,
        ) = ([], [], [], [], [])

        # Iterate over distributions
        for idx, distribution in enumerate(self.distributions):
            # Default cut_off is a static value to reduce errors
            # due to very low decimal numbers
            cut_off = general.MINIMUM_LINE_LENGTH
            if automatic_cut_offs:
                # Use automatic cut off determined by powerlaw
                cut_off = distribution.automatic_fit.xmin
            elif self.cut_offs is not None:
                # Use manual cut offs given as inputs
                cut_off = self.cut_offs[idx]

            assert isinstance(cut_off, float)

            # Resolve the lengths and ccm
            (
                truncated_length_array,
                ccm_array_normed,
            ) = distribution.generate_distributions(cut_off=cut_off)
            (
                full_length_array,
                full_ccm_array_normed,
            ) = distribution.generate_distributions(cut_off=general.MINIMUM_LINE_LENGTH)

            # Collect
            truncated_length_array_all.append(truncated_length_array)
            ccm_array_normed_all.append(ccm_array_normed)
            full_length_array_all.append(full_length_array)
            full_ccm_array_normed_all.append(full_ccm_array_normed)

            cut_offs.append(cut_off)

        # Return all length arrays and all ccm arrays
        return (
            truncated_length_array_all,
            ccm_array_normed_all,
            full_length_array_all,
            full_ccm_array_normed_all,
            cut_offs,
        )

    @property
    def names(self) -> List[str]:
        """
        Get length distribution names.
        """
        return [ld.name for ld in self.distributions]

    def plot_multi_length_distributions(
        self,
        automatic_cut_offs: bool,
        plot_truncated_data: bool,
        scorer: Callable[[np.ndarray, np.ndarray], float] = sklm.mean_squared_log_error,
    ) -> Tuple[Polyfit, Figure, Axes]:
        """
        Plot multi-scale length distribution.
        """
        # Get length arrays and ccm arrays
        (
            truncated_length_array_all,
            ccm_array_normed_all,
            full_length_array_all,
            full_ccm_array_normed_all,
            cut_offs,
        ) = self.normalized_distributions(automatic_cut_offs=automatic_cut_offs)

        # Concatenate
        lengths_concat = np.concatenate(truncated_length_array_all)
        ccm_concat = np.concatenate(ccm_array_normed_all)
        # full_lengths_concat = np.concatenate(full_length_array_all)
        # full_ccm_concat = np.concatenate(full_ccm_array_normed_all)

        # Fit a powerlaw to the multi dataset values
        polyfit = fit_to_multi_scale_lengths(
            ccm=ccm_concat, lengths=lengths_concat, fitter=self.fitter, scorer=scorer
        )
        fig, ax = plot_multi_distributions_and_fit(
            truncated_length_array_all=truncated_length_array_all,
            ccm_array_normed_all=ccm_array_normed_all,
            full_length_array_all=full_length_array_all,
            full_ccm_array_normed_all=full_ccm_array_normed_all,
            cut_offs=cut_offs,
            names=self.names,
            polyfit=polyfit,
            using_branches=self.using_branches,
            plot_truncated_data=plot_truncated_data,
        )
        return polyfit, fig, ax

    def optimize_cut_offs(
        self,
        shgo_kwargs: Optional[Dict[str, Any]] = None,
        scorer: Callable[[np.ndarray, np.ndarray], float] = sklm.mean_squared_log_error,
    ) -> Tuple[MultiScaleOptimizationResult, "MultiLengthDistribution"]:
        """
        Get cut-off optimized MultiLengthDistribution.
        """
        opt_result = self.optimized_multi_scale_fit(
            shgo_kwargs=shgo_kwargs if shgo_kwargs is not None else {}, scorer=scorer
        )
        optimized_mld = MultiLengthDistribution(
            distributions=self.distributions,
            using_branches=self.using_branches,
            cut_offs=list(opt_result.cut_offs),
            _optimized=True,
            fitter=self.fitter,
        )
        return opt_result, optimized_mld

    def optimized_multi_scale_fit(
        self,
        scorer: Callable[[np.ndarray, np.ndarray], float],
        shgo_kwargs: Dict[str, Any],
    ) -> MultiScaleOptimizationResult:
        """
        Use scipy.optimize.shgo to optimize fit.
        """
        # Use automatic powerlaw fitting to get first guesses for optimization
        xmins = [ld.automatic_fit.xmin for ld in self.distributions]
        # xmins = [1.0 for ld in self.distributions]

        # Get all arrays of lengths and use them to define
        # lower and upper bounds for cut-offs
        truncated_length_array_all, _, _, _, _ = self.normalized_distributions(
            automatic_cut_offs=False
        )

        # Bound to upper 99 quantile
        bounds = [np.quantile(arr, (0.0, 1.0)) for arr in truncated_length_array_all]
        # bounds = [(0.0, 1.0) for _ in truncated_length_array_all]

        def constrainer(
            cut_offs: np.ndarray, distributions: List[LengthDistribution]
        ) -> float:
            """
            Constrain sample count to 2 or more.
            """
            lens = 0
            for dist, cut_off in zip(distributions, cut_offs):
                ls, _ = dist.generate_distributions(cut_off=cut_off)
                lens += len(ls)

            # sample count should be 2 or more
            return lens - 1.5

        constraints = {
            "type": "ineq",
            "fun": constrainer,
            "args": (self.distributions,),
        }

        # Run the shgo global optimization to find optimal cut-off values
        # that result in the minimum value of score (loss function).
        # Values will be stored in the x attribute of optimize_result
        optimize_result = shgo(
            optimize_cut_offs,
            args=(self.distributions, self.fitter, scorer),
            bounds=bounds,
            constraints=constraints,
            **shgo_kwargs,
        )
        assert isinstance(optimize_result, OptimizeResult)

        # Use the optimized cut-offs to determine the resulting exponent
        polyfit, proportions = _optimize_cut_offs(
            optimize_result.x,
            distributions=self.distributions,
            fitter=self.fitter,
            scorer=scorer,
        )

        # Return a composed object of the results for easier parsing downstream
        return MultiScaleOptimizationResult(
            polyfit=polyfit,
            cut_offs=optimize_result.x,
            optimize_result=optimize_result,
            x0=np.array(xmins),
            bounds=np.array(bounds),
            proportions_of_data=proportions,
        )


@general.JOBLIB_CACHE.cache
def determine_fit(
    length_array: np.ndarray, cut_off: Optional[float] = None
) -> Optional[powerlaw.Fit]:
    """
    Determine powerlaw (along other) length distribution fits for given data.
    """
    if len(length_array) == 0:
        return None
    length_array_max = length_array.max()
    if cut_off is not None and cut_off >= length_array_max:
        raise ValueError(
            f"Expected lower cut_off ({cut_off}) "
            f"than max of length_array ({length_array_max})."
        )
    fit = (
        # Use cut-off if it is given
        SilentFit(length_array, xmin=cut_off, verbose=False)
        # Otherwise powerlaw determines it automatically using the data
        if cut_off is not None
        else SilentFit(length_array, verbose=False)
    )
    return fit


def plot_fit_on_ax(
    ax: Axes,
    fit: powerlaw.Fit,
    fit_distribution: Dist,
    use_probability_density_function: bool,
) -> None:
    """
    Plot powerlaw model to ax.
    """
    plot_func = "plot_ccdf" if not use_probability_density_function else "plot_pdf"
    if fit_distribution == Dist.POWERLAW:
        getattr(fit.power_law, plot_func)(
            ax=ax, label="Powerlaw", linestyle="--", color="red"
        )
    elif fit_distribution == Dist.LOGNORMAL:
        getattr(fit.lognormal, plot_func)(
            ax=ax, label="Lognormal", linestyle="--", color="lime"
        )
    elif fit_distribution == Dist.EXPONENTIAL:
        getattr(fit.exponential, plot_func)(
            ax=ax, label="Exponential", linestyle="--", color="blue"
        )
    else:
        raise ValueError(f"Expected fit_distribution to be one of {list(Dist)}")


def _setup_length_plot_axlims(
    ax: Axes,
    length_array: np.ndarray,
    ccm_array: np.ndarray,
    # cut_off: float,
):
    """
    Set ax limits for length plotting.
    """
    # truncated_length_array = (
    #     length_array[length_array > cut_off] if cut_off is not None else length_array
    # )

    # TODO: Anomalous very low value lengths can mess up xlims
    left = np.quantile(length_array, 0.01) / 10
    # left = length_array.min() / 10
    right = length_array.max() * 10
    bottom = ccm_array.min() / 20
    top = ccm_array.max() * 100
    try:
        ax.set_xlim(left, right)
        ax.set_ylim(bottom, top)
    except ValueError:
        log.error("Failed to set up x and y limits.", exc_info=True)
        # Don't try setting if it errors


def plot_distribution_fits(
    length_array: np.ndarray,
    label: str,
    using_branches: bool,
    use_probability_density_function: bool,
    cut_off: Optional[float] = None,
    fit: Optional[powerlaw.Fit] = None,
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    fits_to_plot: Tuple[Dist, ...] = (Dist.EXPONENTIAL, Dist.LOGNORMAL, Dist.POWERLAW),
    plain: bool = False,
) -> Tuple[Optional[powerlaw.Fit], Figure, Axes]:
    """
    Plot length distribution and `powerlaw` fits.

    If a powerlaw.Fit is not given it will be automatically determined (using
    the optionally given cut_off).
    """
    if fig is None:
        if ax is None:
            # Create figure, ax
            fig, ax = plt.subplots(figsize=(7, 7))
        else:
            fig_maybe = ax.get_figure()
            assert isinstance(fig_maybe, Figure)
            fig = fig_maybe

    assert isinstance(fig, Figure)
    assert isinstance(ax, Axes)
    assert ax is not None

    if fit is None:
        # Determine powerlaw, exponential, lognormal fits
        fit = determine_fit(length_array, cut_off)

    if fit is None:
        log.warning(
            f"No length data for plotting with label {label}. Returning fit as None."
        )
        return None, fig, ax

    # Get the x, y data from fit
    # y values are either the complementary cumulative distribution function
    # or the probability density function
    # Depends on use_probability_density_function boolean
    cut_off_is_lower = fit.xmin < length_array.min()
    if not use_probability_density_function:
        # complementary cumulative distribution
        truncated_length_array, y_array = fit.ccdf()
        full_length_array, full_y_array = fit.ccdf(original_data=True)
    else:
        # probability density function
        bin_edges, y_array = fit.pdf()

        # Use same bin_edges and y_array if xmin/cut-off is lower than
        # smallest line length
        full_bin_edges, full_y_array = fit.pdf(original_data=True)

        # fit.pdf returns the bin edges. These need to be transformed to
        # centers for plotting.
        truncated_length_array = (bin_edges[:-1] + bin_edges[1:]) / 2.0
        full_length_array = (full_bin_edges[:-1] + full_bin_edges[1:]) / 2.0

        # filter out zeros
        is_zero_y = np.isclose(y_array, 0.0)
        is_zero_full_y = np.isclose(full_y_array, 0.0)
        truncated_length_array = truncated_length_array[~is_zero_y]
        full_length_array = full_length_array[~is_zero_full_y]
        y_array = y_array[~is_zero_y]
        full_y_array = full_y_array[~is_zero_full_y]

    assert len(truncated_length_array) == len(y_array)
    assert len(full_length_array) == len(full_y_array)

    # Plot truncated length scatter plot
    ax.scatter(
        x=truncated_length_array,
        y=y_array,
        s=25,
        label=label,
        alpha=1.0,
        color="black",
        marker="x",
    )

    if not cut_off_is_lower:
        # Normalize full_ccm_array to the truncated ccm_array
        full_y_array = full_y_array / (
            full_y_array[len(full_y_array) - len(y_array)] / y_array.max()
        )
        # Plot full length scatter plot with different color and transparency
        ax.scatter(
            x=full_length_array,
            y=full_y_array,
            s=3,
            # label=f"{label} (cut)",
            alpha=0.5,
            color="gray",
            marker="x",
            zorder=-10,
        )

    # Plot the actual fits (powerlaw, exp...)
    for fit_distribution in fits_to_plot:
        plot_fit_on_ax(
            ax,
            fit,
            fit_distribution,
            use_probability_density_function=use_probability_density_function,
        )

    # Plot cut-off if applicable
    plot_axvline = cut_off is None or cut_off != 0.0
    if plot_axvline and not plain:
        # Indicate cut-off if cut-off is not given as 0.0
        truncated_length_array_min = min((truncated_length_array.min(), fit.xmin))
        ax.axvline(
            truncated_length_array_min,
            linestyle="dotted",
            color="black",
            alpha=0.6,
            label="Cut-Off",
            linewidth=3.5,
        )
        ax.text(
            truncated_length_array_min,
            y_array.min(),
            f"{round(truncated_length_array_min, 2)} m",
            rotation=90,
            horizontalalignment="right",
            fontsize="xx-large",
        )

    # Set title with exponent
    rounded_exponent = round(calculate_exponent(fit.alpha), 3)
    target = "Branches" if using_branches else "Traces"
    if not plain:
        ax.set_title(
            f"{label}\nPower-law Exponent ({target}) = ${rounded_exponent}$",
            fontdict=dict(fontsize="xx-large"),
        )

    # Setup of ax appearance and axlims
    setup_ax_for_ld(
        ax,
        using_branches=using_branches,
        indiv_fit=True,
        use_probability_density_function=use_probability_density_function,
        plain=plain,
    )
    _setup_length_plot_axlims(
        ax=ax,
        length_array=truncated_length_array,
        ccm_array=y_array,
    )

    return fit, fig, ax


def setup_length_dist_legend(ax: Axes):
    """
    Set up legend for length distribution plots.

    Used for both single and multi distribution plots.
    """
    # Setup legend
    handles, labels = ax.get_legend_handles_labels()
    labels = [fill(label, 16) for label in labels]
    lgnd = ax.legend(
        handles,
        labels,
        loc="upper right",
        # bbox_to_anchor=(1.37, 1.02),
        ncol=2,
        columnspacing=0.3,
        # shadow=True,
        prop={"size": "xx-large"},
        framealpha=0.6,
        facecolor="white",
        # fontsize=40,
    )

    # Setup legend line widths larger
    # TODO: This can be cleaned to use legend_handles only when matplotlib min
    # version is high enough
    legend_handles = getattr(lgnd, "legend_handles", None)
    legend_handles = (
        legend_handles if legend_handles is not None else lgnd.legendHandles
    )
    for lh in legend_handles:
        # lh._sizes = [750]
        lh.set_linewidth(3)


def setup_ax_for_ld(
    ax: Axes,
    using_branches: bool,
    indiv_fit: bool,
    use_probability_density_function: bool,
    plain: bool = False,
):
    """
    Configure ax for length distribution plots.

    :param ax: Ax to setup.
    :param using_branches: Are the lines in the axis branches or traces.
    :param indiv_fit: Is the plot single-scale or multi-scale.
    :param use_probability_density_function: Whether to use complementary
           cumulative distribution function
    :param plain: Should the stylizing be kept to a minimum.
    """
    # LABELS
    label = "Branch Length $(m)$" if using_branches else "Trace Length $(m)$"
    font_props = dict(
        fontsize="xx-large",
        fontfamily="DejaVu Sans",
        style="italic",
        fontweight="bold",
    )
    ax.set_xlabel(
        label,
        labelpad=14,
        **font_props,
    ) if not plain else None
    # Individual powerlaw fits are not normalized to area because they aren't
    # multiscale
    ccm_unit = r"$(\frac{1}{m^2})$" if not indiv_fit else ""
    prefix = "" if indiv_fit else "AN"
    function_name = "CCM" if not use_probability_density_function else "PDF"
    ax.set_ylabel(
        f"{prefix}{function_name} {ccm_unit}",
        # prefix + function_name + ccm_unit,
        **font_props,
    ) if not plain else None

    # Setup x and y axis ticks and their labels
    ax.xaxis.set_ticks(
        ticks=ax.xaxis.get_ticklocs(),
        labels=ax.xaxis.get_ticklabels(),
        color="black",
        fontsize="x-large",
    )
    ax.yaxis.set_ticks(
        ticks=ax.yaxis.get_ticklocs(),
        labels=ax.yaxis.get_ticklabels(),
        color="black",
        fontsize="x-large",
    )
    ax.tick_params(axis="both", width=1.2)

    # Setup legend
    setup_length_dist_legend(ax=ax) if not plain else None

    # Setup grid
    ax.grid(zorder=-10, color="black", alpha=0.25)

    # Change x and y scales to logarithmic
    ax.set_xscale("log")
    ax.set_yscale("log")

    # Set facecolor depending on using_branches
    # facecolor = "oldlace" if using_branches else "gray"
    # ax.set_facecolor(facecolor)


def distribution_compare_dict(fit: powerlaw.Fit) -> Dict[str, float]:
    """
    Compose a dict of length distribution fit comparisons.
    """
    compare_dict = dict()
    for dist_enum_pairs in [
        (Dist.POWERLAW, Dist.LOGNORMAL),
        (Dist.POWERLAW, Dist.EXPONENTIAL),
        (Dist.LOGNORMAL, Dist.EXPONENTIAL),
        (Dist.POWERLAW, Dist.TRUNCATED_POWERLAW),
    ]:
        first, second = dist_enum_pairs[0].value, dist_enum_pairs[1].value
        r, p = fit.distribution_compare(first, second, normalized_ratio=True)
        compare_dict[f"{first} vs. {second} R"] = r
        compare_dict[f"{first} vs. {second} p"] = p
    return compare_dict


def all_fit_attributes_dict(fit: powerlaw.Fit) -> Dict[str, float]:
    """
    Collect 'all' fit attributes into a dict.
    """
    return {
        # **describe_powerlaw_fit(fit),
        # Attributes for remaking fits
        Dist.LOGNORMAL.value + " " + SIGMA: fit.lognormal.sigma,
        Dist.LOGNORMAL.value + " " + MU: fit.lognormal.mu,
        Dist.EXPONENTIAL.value + " " + LAMBDA: fit.exponential.Lambda,
        Dist.TRUNCATED_POWERLAW.value + " " + LAMBDA: fit.truncated_power_law.Lambda,
        Dist.TRUNCATED_POWERLAW.value + " " + ALPHA: fit.truncated_power_law.alpha,
        Dist.TRUNCATED_POWERLAW.value + " "
        # + EXPONENT: -(fit.truncated_power_law.alpha - 1),
        + EXPONENT: calculate_exponent(fit.truncated_power_law.alpha),
        # Fit statistics
        Dist.LOGNORMAL.value + " " + LOGLIKELIHOOD: fit.lognormal.loglikelihood,
        Dist.EXPONENTIAL.value + " " + LOGLIKELIHOOD: fit.exponential.loglikelihood,
        Dist.TRUNCATED_POWERLAW.value
        + " "
        + LOGLIKELIHOOD: fit.truncated_power_law.loglikelihood,
    }


def calculate_exponent(alpha: float):
    """
    Calculate exponent from powerlaw.alpha.
    """
    return -(alpha - 1)


def cut_off_proportion_of_data(fit: powerlaw.Fit, length_array: np.ndarray) -> float:
    """
    Get the proportion of data cut off by `powerlaw` cut off.

    If no fit is passed the cut off is the one used in `automatic_fit`.
    """
    arr_less_than = length_array < fit.xmin
    assert isinstance(arr_less_than, np.ndarray)
    return sum(arr_less_than) / len(length_array) if len(length_array) > 0 else 0.0


def describe_powerlaw_fit(
    fit: powerlaw.Fit, length_array: np.ndarray, label: Optional[str] = None
) -> Dict[str, float]:
    """
    Compose dict of fit powerlaw attributes and comparisons between fits.
    """
    base = {
        **distribution_compare_dict(fit),
        Dist.POWERLAW.value + " " + KOLM_DIST: fit.power_law.D,
        Dist.EXPONENTIAL.value + " " + KOLM_DIST: fit.exponential.D,
        Dist.LOGNORMAL.value + " " + KOLM_DIST: fit.lognormal.D,
        Dist.TRUNCATED_POWERLAW.value + " " + KOLM_DIST: fit.truncated_power_law.D,
        Dist.POWERLAW.value + " " + ALPHA: fit.alpha,
        Dist.POWERLAW.value + " " + EXPONENT: calculate_exponent(fit.alpha),
        Dist.POWERLAW.value + " " + CUT_OFF: fit.xmin,
        Dist.POWERLAW.value + " " + SIGMA: fit.power_law.sigma,
        **all_fit_attributes_dict(fit),
        "lengths cut off proportion": cut_off_proportion_of_data(
            fit=fit, length_array=length_array
        ),
    }
    if label is None:
        return base
    return {f"{label} {key}": value for key, value in base.items()}


def sort_and_log_lengths_and_ccm(
    lengths: np.ndarray, ccm: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Preprocess lengths and ccm.

    Sorts them and calculates their natural logarithmic.
    """
    # Find sorted order of lengths
    length_sort_permutation = lengths.argsort()

    # Apply sorted order of lengths to both lengths and ccms
    log_lengths_sorted = np.log(lengths[length_sort_permutation])
    log_ccm_sorted = np.log(ccm[length_sort_permutation])

    return log_lengths_sorted, log_ccm_sorted


def calculate_fitted_values(
    log_lengths: np.ndarray, m_value: float, constant: float
) -> np.ndarray:
    """
    Calculate fitted values of y.
    """
    # Calculate the fitted values of y
    y_fit = np.exp(m_value * log_lengths + constant)

    return y_fit


def fit_to_multi_scale_lengths(
    ccm: np.ndarray,
    lengths: np.ndarray,
    fitter: Callable[[np.ndarray, np.ndarray], Tuple[float, float]] = numpy_polyfit,
    scorer: Callable[[np.ndarray, np.ndarray], float] = sklm.mean_squared_log_error,
) -> Polyfit:
    """
    Fit np.polyfit to multiscale length distributions.

    Returns the fitted values, exponent and constant of fit
    within a ``Polyfit`` instance.
    """
    log_lengths_sorted, log_ccm_sorted = sort_and_log_lengths_and_ccm(
        lengths=lengths, ccm=ccm
    )

    # Fit numpy polyfit to data
    fit_vals = fitter(log_lengths_sorted, log_ccm_sorted)

    assert len(fit_vals) == 2

    m_value, constant = fit_vals
    log.info(
        "Fitted with fitter.",
        extra=dict(fitter=fitter, m_value=m_value, constant=constant),
    )

    # Calculate the fitted values of y
    y_fit = calculate_fitted_values(
        log_lengths=log_lengths_sorted,
        m_value=m_value,
        constant=constant,
    )
    score = scorer(ccm, y_fit)
    assert isinstance(score, float)

    return Polyfit(
        y_fit=y_fit, m_value=m_value, constant=constant, score=score, scorer=scorer
    )


def plot_multi_distributions_and_fit(
    truncated_length_array_all: List[np.ndarray],
    ccm_array_normed_all: List[np.ndarray],
    full_length_array_all: List[np.ndarray],
    full_ccm_array_normed_all: List[np.ndarray],
    cut_offs: List[np.ndarray],
    names: List[str],
    polyfit: Polyfit,
    using_branches: bool,
    plot_truncated_data: bool,
) -> Tuple[Figure, Axes]:
    """
    Plot multi-scale length distribution.
    """
    # Start making plot
    fig, ax = plt.subplots(figsize=(7, 7))
    assert isinstance(ax, Axes)
    assert isinstance(fig, Figure)

    # Determine scorer name
    # If not found in SCORER_NAMES, the __name__ of the scorer function is used
    try:
        scorer_str = SCORER_NAMES[polyfit.scorer]
    except KeyError:
        log.warning(
            "Expected to find name string for polyfit scorer: "
            f"{polyfit.scorer} in SCORER_NAMES."
        )
        scorer_str = polyfit.scorer.__name__

    # Determine score string (float or scientific notation)
    is_very_low_score = polyfit.score < 0.01
    score_descriptive = (
        f"{polyfit.score:.2e}" if is_very_low_score else str(round(polyfit.score, 2))
    )

    # Set title with powerlaw exponent and score
    ax.set_title(
        (
            f"Exponent = {polyfit.m_value:.2f} and"
            f" Score ({scorer_str}) = {score_descriptive}"
        ),
        fontsize="x-large",
    )

    full_ccm_array_normed_concat = np.concatenate(full_ccm_array_normed_all)

    # Make color cycle
    color_cycle = cycle(sns.color_palette("dark", 5))

    # Cycle for cut-off label placement
    label_cycle = cycle([True, False])
    # Plot length distributions
    included_cut_off_label = False
    for (
        name,
        truncated_length_array,
        ccm_array_normed,
        full_length_array,
        full_ccm_array,
        cut_off,
    ) in zip(
        names,
        truncated_length_array_all,
        ccm_array_normed_all,
        full_length_array_all,
        full_ccm_array_normed_all,
        cut_offs,
    ):
        # Get color for distribution
        point_color = next(color_cycle)

        if len(truncated_length_array) > 0:
            # Indicate the cut-off with a dotted vertical line
            # truncated_length_array_min = truncated_length_array.min()
            cut_off_legend_kwarg = dict(label="Cut-Off(s)")
            ax.axvline(
                cut_off,
                linestyle="dotted",
                color="black",
                alpha=0.6,
                linewidth=3.5,
                **(cut_off_legend_kwarg if not included_cut_off_label else dict()),
            )
            included_cut_off_label = True

            # Also indicate cut-off value, embedded into plot
            plot_low = next(label_cycle)
            ax.text(
                cut_off,
                ccm_array_normed.min() if plot_low else ccm_array_normed.max(),
                f"{round(cut_off, 2)} m",
                rotation=90,
                horizontalalignment="right",
                fontsize="xx-large",
                verticalalignment="center" if plot_low else "bottom",
            )

        # Plot main data, not truncated
        ax.scatter(
            x=truncated_length_array,
            y=ccm_array_normed,
            s=25,
            label=name,
            marker="X",
            color=point_color,
        )
        if plot_truncated_data:
            # Plot truncated data part
            ax.scatter(
                x=full_length_array,
                y=full_ccm_array,
                s=3,
                marker="x",
                color="gray",
                alpha=0.5,
                zorder=-10,
            )

    # Plot polyfit
    y_fit = polyfit.y_fit

    # Concatenate all individually truncated length data and
    # plot the fitted y values for the lengths
    concatted_lengths = np.concatenate(truncated_length_array_all)
    ax.plot(
        (concatted_lengths.max(), concatted_lengths.min()),
        (y_fit.min(), y_fit.max()),
        label="Polyfit",
        linestyle="dashed",
    )

    # Setup axes
    setup_ax_for_ld(
        ax=ax,
        using_branches=using_branches,
        indiv_fit=False,
        use_probability_density_function=False,
    )
    _setup_length_plot_axlims(
        ax=ax,
        length_array=np.concatenate(full_length_array_all),
        ccm_array=full_ccm_array_normed_concat,
    )

    # Add legend
    # setup_length_dist_legend(ax=ax)

    return fig, ax


def _optimize_cut_offs(
    cut_offs: np.ndarray,
    distributions: List[LengthDistribution],
    fitter: Callable[[np.ndarray, np.ndarray], Tuple[float, float]],
    scorer: Callable[[np.ndarray, np.ndarray], float],
) -> Tuple[Polyfit, List[float]]:
    # Each length and associated ccm must be collected
    cut_length_arrays, cut_ccm_arrays, proportions = [], [], []

    # Iterate over given distributions and cut-offs
    for dist, cut_off in zip(distributions, cut_offs):
        # Use LengthDistribution.apply_cut_off to get truncated and normalized
        # length and ccm data
        cut_lengths, cut_ccm = dist.generate_distributions(cut_off=cut_off)
        cut_off_proportion = len(cut_lengths) / len(dist.lengths)
        cut_length_arrays.append(cut_lengths)
        cut_ccm_arrays.append(cut_ccm)
        proportions.append(cut_off_proportion)

    # Concatenate length and ccm data into single arrays
    length_array_concat = np.concatenate(cut_length_arrays)
    ccm_array_concat = np.concatenate(cut_ccm_arrays)

    # Fit a regression fit to the concatenated arrays
    polyfit = fit_to_multi_scale_lengths(
        ccm=ccm_array_concat, lengths=length_array_concat, fitter=fitter, scorer=scorer
    )

    return polyfit, proportions


def optimize_cut_offs(
    cut_offs: np.ndarray,
    distributions: List[LengthDistribution],
    fitter: Callable[[np.ndarray, np.ndarray], Tuple[float, float]],
    scorer: Callable[[np.ndarray, np.ndarray], float],
    # TODO: extra args are passed in conda ci-tests.
    #    def function_wrapper(x, *wrapper_args):
    # ncalls[0] += 1
    # # A copy of x is sent to the user function (gh13740)
    # >       fx = function(np.copy(x), *(wrapper_args + args))
    # E       TypeError: optimize_cut_offs() takes 4 positional arguments but 7 were given
    *_,
) -> float:
    """
    Optimize multiscale fit.

    Requirements for the optimization function are that the function must take
    one argument of 1-d array and return a single float. It can take
    static arguments (distributions, fitter).
    """
    polyfit, _ = _optimize_cut_offs(
        cut_offs=cut_offs,
        distributions=distributions,
        fitter=fitter,
        scorer=scorer,
    )
    return polyfit.score


def scikit_linear_regression(
    log_lengths: np.ndarray, log_ccm: np.ndarray
) -> Tuple[float, float]:
    """
    Fit using scikit LinearRegression.
    """
    # Fit the regression
    model = LinearRegression().fit(log_lengths.reshape((-1, 1)), log_ccm)

    # Get coefficients and intercept (==m_value and constant)
    coefs = model.coef_
    intercept = model.intercept_

    assert len(coefs) == 1
    assert isinstance(intercept, float)
    m_value = coefs[0]
    assert isinstance(m_value, float)

    return m_value, intercept


def sorted_lengths_and_ccm(
    lengths: np.ndarray, area_value: Optional[float]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Get (normalized) complementary cumulative number array.

    Give area_value as None to **not** normalize.

    >>> lengths = np.array([2, 4, 8, 16, 32])
    >>> area_value = 10.0
    >>> sorted_lengths_and_ccm(lengths, area_value)
    (array([ 2,  4,  8, 16, 32]), array([0.1 , 0.08, 0.06, 0.04, 0.02]))

    >>> lengths = np.array([2, 4, 8, 16, 32])
    >>> area_value = None
    >>> sorted_lengths_and_ccm(lengths, area_value)
    (array([ 2,  4,  8, 16, 32]), array([1. , 0.8, 0.6, 0.4, 0.2]))
    """
    # use powerlaw.Fit to determine ccm
    lengths, ccm = SilentFit(data=lengths, xmin=general.MINIMUM_LINE_LENGTH).ccdf(
        original_data=True
    )

    if area_value is not None:
        # Normalize with area
        ccm = ccm / area_value

    return lengths, ccm


def apply_cut_off(
    lengths: np.ndarray, ccm: np.ndarray, cut_off: float = general.MINIMUM_LINE_LENGTH
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Apply cut-off to length data and associated ccm.

    >>> lengths = np.array([2, 4, 8, 16, 32])
    >>> ccm = np.array([1. , 0.8, 0.6, 0.4, 0.2])
    >>> cut_off = 4.5
    >>> apply_cut_off(lengths, ccm, cut_off)
    (array([ 8, 16, 32]), array([0.6, 0.4, 0.2]))
    """
    are_over_cut_off = lengths > cut_off

    return lengths[are_over_cut_off], ccm[are_over_cut_off]
